# Test program for Contractor which exposes a type Text and associated procedures.

# author R.N.Bosworth

# version 27 Feb 2023   15:32

from emily0_9 import texting

"""
Copyright (C) 2012,2015,2016,2017,2018,2019,2021,2022,2023  R.N.Bosworth

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License (gpl.txt) for more details.
"""

# test program
# ------------

print("Tests of texting.new_text",end=' ')
t = texting.new_text()
assert len(t._my_text) == 1
assert t._cursor_line == 0
assert t._cursor_char == 0
assert t._modified == False
assert t._first_paragraph == True
t = texting.new_text()
assert len(t._my_text) == 1
assert t._cursor_line == 0
assert t._cursor_char == 0
assert t._modified == False
assert t._first_paragraph == True  
print("OK")

print("Tests of texting.advance", end=' ')
#texting.advance(None)
t = texting.new_text()
assert texting.cursor_line_offset(t) == 0
assert texting.cursor_code_point_offset(t) == 0
texting.advance(t)
assert texting.cursor_line_offset(t) == 0
assert texting.cursor_code_point_offset(t) == 0
t = texting.new_text()
texting.insert_code_point(t,ord('a'))
texting.insert_code_point(t,ord('\n'))
texting.insert_code_point(t,ord('b'))
texting.set_cursor_start(t)
texting.advance(t)
assert texting.cursor_line_offset(t) == 0
assert texting.cursor_code_point_offset(t) == 1
texting.advance(t)
assert texting.cursor_line_offset(t) == 1
assert texting.cursor_code_point_offset(t) == 0
texting.advance(t)
assert texting.cursor_line_offset(t) == 1
assert texting.cursor_code_point_offset(t) == 1
texting.advance(t)
assert texting.cursor_line_offset(t) == 1
assert texting.cursor_code_point_offset(t) == 1
print("OK")

print("Tests of texting.current_code_point", end=' ')
#texting.current_code_point(None)
t = texting.new_text()
assert texting.current_code_point(t) == texting.END_OF_TEXT
t = texting.new_text()
texting.insert_code_point(t,ord('a'))
texting.insert_code_point(t,ord('\n'))
texting.insert_code_point(t,ord('b'))
texting.set_cursor_start(t)
assert texting.current_code_point(t) == ord('a')
assert texting.current_code_point(t) == ord('a')
texting.advance(t)
assert texting.current_code_point(t) == ord('\n')
texting.advance(t)
assert texting.current_code_point(t) == ord('b')
texting.advance(t)
assert texting.current_code_point(t) == texting.END_OF_TEXT
texting.advance(t)
assert texting.current_code_point(t) == texting.END_OF_TEXT
print("OK")

print("Tests of texting.get_alignment, texting.set_alignment", end=' ')
#texting.get_alignment(None)
t = texting.new_text()
assert texting.get_alignment(t) == texting.Alignment.BEGIN
#texting.set_alignment(None,None)
#texting.set_alignment(t,None)
texting.set_alignment(t,texting.Alignment.MIDDLE)
assert texting.get_alignment(t) == texting.Alignment.MIDDLE
texting.insert_code_point(t,ord('\n'))
assert texting.get_alignment(t) == texting.Alignment.MIDDLE
texting.set_alignment(t,texting.Alignment.END)
assert texting.get_alignment(t) == texting.Alignment.END
print("OK")

print("Tests of texting.insert_code_point", end=' ')
#texting.insert_code_point(None,None)
t = texting.new_text()
#texting.insert_code_point(t,None)
#texting.insert_code_point(t,-1)
#texting.insert_code_point(t,0x110000)
assert texting.has_been_modified(t) == False
texting.set_cursor(t,0,0)
assert texting.get_alignment(t) == texting.Alignment.BEGIN
assert texting.cursor_line_offset(t) == 0
assert texting.cursor_code_point_offset(t) == 0
texting.insert_code_point(t,ord('\n'))
assert texting.get_alignment(t) == texting.Alignment.BEGIN
assert texting.cursor_line_offset(t) == 1
assert texting.cursor_code_point_offset(t) == 0
assert texting.has_been_modified(t) == True
texting.set_cursor(t,0,0)
assert texting.get_alignment(t) == texting.Alignment.BEGIN
assert texting.current_code_point(t) == ord('\n')
texting.advance(t)
assert texting.current_code_point(t) == texting.END_OF_TEXT
t = texting.new_text()
assert texting.has_been_modified(t) == False
texting.set_cursor(t,0,0)
texting.insert_code_point(t,ord('a'))
assert texting.cursor_line_offset(t) == 0
assert texting.cursor_code_point_offset(t) == 1
assert texting.has_been_modified(t) == True
texting.set_cursor(t,0,0)
assert texting.current_code_point(t) == ord('a')
texting.advance(t)
assert texting.current_code_point(t) == texting.END_OF_TEXT
t = texting.new_text()
texting.set_alignment(t,texting.Alignment.MIDDLE)
texting.insert_code_point(t,ord('a'))
texting.insert_code_point(t,ord('b'))
texting.insert_code_point(t,ord('c'))
texting.insert_code_point(t,ord('d'))
texting.insert_code_point(t,ord('\n'))
texting.insert_code_point(t,ord('e'))
texting.insert_code_point(t,ord('f'))
assert texting.has_been_modified(t) == True
texting.set_cursor(t,0,0)
assert texting.get_alignment(t) == texting.Alignment.MIDDLE
assert texting.current_code_point(t) == ord('a')
texting.advance(t)
assert texting.current_code_point(t) == ord('b')
texting.advance(t)
assert texting.current_code_point(t) == ord('c')
texting.advance(t)
assert texting.current_code_point(t) == ord('d')
texting.advance(t)
assert texting.current_code_point(t) == ord('\n')
texting.advance(t)
assert texting.get_alignment(t) == texting.Alignment.MIDDLE
assert texting.current_code_point(t) == ord('e')
texting.advance(t)
assert texting.current_code_point(t) == ord('f')
texting.advance(t)
assert texting.current_code_point(t) == texting.END_OF_TEXT
texting.set_cursor(t,0,0)
texting.insert_code_point(t,ord('\n'))
texting.set_cursor(t,0,0)
assert texting.get_alignment(t) == texting.Alignment.MIDDLE
assert texting.current_code_point(t) == ord('\n')
texting.advance(t)
assert texting.get_alignment(t) == texting.Alignment.MIDDLE
assert texting.current_code_point(t) == ord('a')
texting.advance(t)
assert texting.current_code_point(t) == ord('b')
texting.advance(t)
assert texting.current_code_point(t) == ord('c')
texting.advance(t)
assert texting.current_code_point(t) == ord('d')
texting.advance(t)
assert texting.current_code_point(t) == ord('\n')
texting.advance(t)
assert texting.get_alignment(t) == texting.Alignment.MIDDLE
assert texting.current_code_point(t) == ord('e')
texting.advance(t)
assert texting.current_code_point(t) == ord('f')
texting.advance(t)
assert texting.current_code_point(t) == texting.END_OF_TEXT
t = texting.new_text()
texting.set_alignment(t,texting.Alignment.MIDDLE)
texting.insert_code_point(t,ord('a'))
texting.insert_code_point(t,ord('b'))
texting.insert_code_point(t,ord('c'))
texting.insert_code_point(t,ord('d'))
texting.insert_code_point(t,ord('\n'))
texting.insert_code_point(t,ord('e'))
texting.insert_code_point(t,ord('f'))
texting.set_cursor(t,0,4)
assert texting.get_alignment(t) == texting.Alignment.MIDDLE
texting.insert_code_point(t,ord('\n'))
texting.set_cursor(t,0,0)
assert texting.get_alignment(t) == texting.Alignment.MIDDLE
assert texting.current_code_point(t) == ord('a')
texting.advance(t)
assert texting.current_code_point(t) == ord('b')
texting.advance(t)
assert texting.current_code_point(t) == ord('c')
texting.advance(t)
assert texting.current_code_point(t) == ord('d')
texting.advance(t)
assert texting.current_code_point(t) == ord('\n')
texting.advance(t)
assert texting.get_alignment(t) == texting.Alignment.MIDDLE
assert texting.current_code_point(t) == ord('\n')
texting.advance(t)
assert texting.get_alignment(t) == texting.Alignment.MIDDLE
assert texting.current_code_point(t) == ord('e')
texting.advance(t)
assert texting.current_code_point(t) == ord('f')
texting.advance(t)
assert texting.current_code_point(t) == texting.END_OF_TEXT
t = texting.new_text()
texting.set_alignment(t,texting.Alignment.MIDDLE)
texting.insert_code_point(t,ord('a'))
texting.insert_code_point(t,ord('b'))
texting.insert_code_point(t,ord('c'))
texting.insert_code_point(t,ord('d'))
texting.insert_code_point(t,ord('\n'))
texting.insert_code_point(t,ord('e'))
texting.insert_code_point(t,ord('f'))
texting.set_cursor(t,0,2)
assert texting.get_alignment(t) == texting.Alignment.MIDDLE
texting.insert_code_point(t,ord('\n'))
texting.set_cursor(t,0,0)
assert texting.get_alignment(t) == texting.Alignment.MIDDLE
assert texting.current_code_point(t) == ord('a')
texting.advance(t)
assert texting.current_code_point(t) == ord('b')
texting.advance(t)
assert texting.current_code_point(t) == ord('\n')
texting.advance(t)
assert texting.get_alignment(t) == texting.Alignment.MIDDLE
assert texting.current_code_point(t) == ord('c')
texting.advance(t)
assert texting.current_code_point(t) == ord('d')
texting.advance(t)
assert texting.current_code_point(t) == ord('\n')
texting.advance(t)
assert texting.get_alignment(t) == texting.Alignment.MIDDLE
assert texting.current_code_point(t) == ord('e')
texting.advance(t)
assert texting.current_code_point(t) == ord('f')
texting.advance(t)
assert texting.current_code_point(t) == texting.END_OF_TEXT
t = texting.new_text()
texting.set_alignment(t,texting.Alignment.MIDDLE)
texting.insert_code_point(t,ord('a'))
texting.insert_code_point(t,ord('b'))
texting.insert_code_point(t,ord('c'))
texting.insert_code_point(t,ord('d'))
texting.insert_code_point(t,ord('\n'))
texting.insert_code_point(t,ord('e'))
texting.insert_code_point(t,ord('f'))
texting.set_cursor(t,0,0)
assert texting.get_alignment(t) == texting.Alignment.MIDDLE
texting.insert_code_point(t,ord('a'))
texting.set_cursor(t,0,0)
assert texting.get_alignment(t) == texting.Alignment.MIDDLE
assert texting.current_code_point(t) == ord('a')
texting.advance(t)
assert texting.current_code_point(t) == ord('a')
texting.advance(t)
assert texting.current_code_point(t) == ord('b')
texting.advance(t)
assert texting.current_code_point(t) == ord('c')
texting.advance(t)
assert texting.current_code_point(t) == ord('d')
texting.advance(t)
assert texting.current_code_point(t) == ord('\n')
texting.advance(t)
assert texting.get_alignment(t) == texting.Alignment.MIDDLE
assert texting.current_code_point(t) == ord('e')
texting.advance(t)
assert texting.current_code_point(t) == ord('f')
texting.advance(t)
assert texting.current_code_point(t) == texting.END_OF_TEXT
t = texting.new_text()
texting.set_alignment(t,texting.Alignment.MIDDLE)
texting.insert_code_point(t,ord('a'))
texting.insert_code_point(t,ord('b'))
texting.insert_code_point(t,ord('c'))
texting.insert_code_point(t,ord('d'))
texting.insert_code_point(t,ord('\n'))
texting.insert_code_point(t,ord('e'))
texting.insert_code_point(t,ord('f'))
texting.set_cursor(t,0,4)
texting.insert_code_point(t,ord('a'))
texting.set_cursor(t,0,0)
assert texting.get_alignment(t) == texting.Alignment.MIDDLE
assert texting.current_code_point(t) == ord('a')
texting.advance(t)
assert texting.current_code_point(t) == ord('b')
texting.advance(t)
assert texting.current_code_point(t) == ord('c')
texting.advance(t)
assert texting.current_code_point(t) == ord('d')
texting.advance(t)
assert texting.current_code_point(t) == ord('a')
texting.advance(t)
assert texting.current_code_point(t) == ord('\n')
texting.advance(t)
assert texting.get_alignment(t) == texting.Alignment.MIDDLE
assert texting.current_code_point(t) == ord('e')
texting.advance(t)
assert texting.current_code_point(t) == ord('f')
texting.advance(t)
assert texting.current_code_point(t) == texting.END_OF_TEXT
t = texting.new_text()
texting.set_alignment(t,texting.Alignment.MIDDLE)
texting.insert_code_point(t,ord('a'))
texting.insert_code_point(t,ord('b'))
texting.insert_code_point(t,ord('c'))
texting.insert_code_point(t,ord('d'))
texting.insert_code_point(t,ord('\n'))
texting.insert_code_point(t,ord('e'))
texting.insert_code_point(t,ord('f'))
texting.set_cursor(t,0,2)
assert texting.get_alignment(t) == texting.Alignment.MIDDLE
texting.insert_code_point(t,ord('a'))
texting.set_cursor(t,0,0)
assert texting.get_alignment(t) == texting.Alignment.MIDDLE
assert texting.current_code_point(t) == ord('a')
texting.advance(t)
assert texting.current_code_point(t) == ord('b')
texting.advance(t)
assert texting.current_code_point(t) == ord('a')
texting.advance(t)
assert texting.current_code_point(t) == ord('c')
texting.advance(t)
assert texting.current_code_point(t) == ord('d')
texting.advance(t)
assert texting.current_code_point(t) == ord('\n')
texting.advance(t)
assert texting.get_alignment(t) == texting.Alignment.MIDDLE
assert texting.current_code_point(t) == ord('e')
texting.advance(t)
assert texting.current_code_point(t) == ord('f')
texting.advance(t)
assert texting.current_code_point(t) == texting.END_OF_TEXT
print("OK")

print("Tests of texting.set_unmodified, texting.set_modified, texting.has_been_modified", end=' ')
#texting.has_been_modified(None)
t = texting.new_text()
assert texting.has_been_modified(t) == False
texting.set_cursor(t,0,0)
texting.insert_code_point(t,ord('a'))
assert texting.has_been_modified(t) == True
#texting.set_unmodified(None)
texting.set_unmodified(t)
assert texting.has_been_modified(t) == False
#texting.set_modified(None)
texting.set_modified(t)
assert texting.has_been_modified(t) == True
print("OK")

print("Tests of texting._line_length", end=' ')
t = texting.new_text()
assert texting._line_length(t) == 0
texting.insert_code_point(t,ord('a'))
texting.insert_code_point(t,ord('b'))
texting.insert_code_point(t,ord('c'))
texting.insert_code_point(t,ord('\n'))
assert texting._line_length(t) == 0
texting.set_cursor(t,0,2)
assert texting._line_length(t) == 3
print("OK")

print("Tests of texting.set_cursor, texting.cursor_code_point_offset, texting.cursor_line_offset", end=' ')
#texting.set_cursor(None,None,None)
t = texting.new_text()
#texting.set_cursor(t,None,None)
#texting.set_cursor(t,-1,None)
#texting.set_cursor(t,0,None)
#texting.set_cursor(t,0,-1)
texting.set_cursor(t,0,0)
texting.insert_code_point(t,ord('a'))
texting.insert_code_point(t,ord('b'))
texting.insert_code_point(t,ord('c'))
texting.insert_code_point(t,ord('\n'))
texting.insert_code_point(t,ord('d'))
texting.insert_code_point(t,ord('e'))
texting.insert_code_point(t,ord('f'))
#texting.set_cursor(t,2,4)
#texting.set_cursor(t,1,4)
texting.set_cursor(t,1,3)
#assert texting.cursor_line_offset(None) == 1
assert texting.cursor_line_offset(t) == 1
#assert texting.cursor_code_point_offset(None) == 3
assert texting.cursor_code_point_offset(t) == 3
t2 = texting.new_text()
texting.set_cursor(t2,0,0)
assert texting.cursor_line_offset(t2) == 0
assert texting.cursor_code_point_offset(t2) == 0
assert texting.cursor_line_offset(t) == 1
assert texting.cursor_code_point_offset(t) == 3
print("OK")

print("Tests of texting.retreat", end=' ')
#texting.retreat(None)
t = texting.new_text()
assert texting.cursor_line_offset(t) == 0
assert texting.cursor_code_point_offset(t) == 0
texting.retreat(t)
assert texting.cursor_line_offset(t) == 0
assert texting.cursor_code_point_offset(t) == 0
t = texting.new_text()
texting.insert_code_point(t,ord('a'))
texting.insert_code_point(t,ord('\n'))
texting.insert_code_point(t,ord('b'))
texting.set_cursor_end(t)
assert texting.cursor_line_offset(t) == 1
assert texting.cursor_code_point_offset(t) == 1
texting.retreat(t)
assert texting.cursor_line_offset(t) == 1
assert texting.cursor_code_point_offset(t) == 0
texting.retreat(t)
assert texting.cursor_line_offset(t) == 0
assert texting.cursor_code_point_offset(t) == 1
texting.retreat(t)
assert texting.cursor_line_offset(t) == 0
assert texting.cursor_code_point_offset(t) == 0
texting.retreat(t)
assert texting.cursor_line_offset(t) == 0
assert texting.cursor_code_point_offset(t) == 0
print("OK")

print("Tests of texting.set_cursor_start, texting.set_cursor_end", end=' ')
#texting.set_cursor_start(None)
#texting.set_cursor_end(None)
t = texting.new_text()
texting.insert_code_point(t,ord('a'))
texting.insert_code_point(t,ord('b'))
texting.insert_code_point(t,ord('c'))
texting.insert_code_point(t,ord('\n'))
texting.insert_code_point(t,ord('d'))
texting.insert_code_point(t,ord('e'))
assert texting.cursor_line_offset(t) == 1
assert texting.cursor_code_point_offset(t) == 2
texting.set_cursor_start(t)
assert texting.cursor_line_offset(t) == 0
assert texting.cursor_code_point_offset(t) == 0
texting.set_cursor_end(t)
assert texting.cursor_line_offset(t) == 1
assert texting.cursor_code_point_offset(t) == 2
print("OK")

print("Tests of texting.delete_after", end=' ')
#texting.delete_after(None)
t = texting.new_text()
assert texting.has_been_modified(t) == False
texting.set_cursor_start(t)
texting.insert_code_point(t,ord('a'))
texting.insert_code_point(t,ord('b'))
texting.insert_code_point(t,ord('c'))
texting.insert_code_point(t,ord('\n'))
texting.insert_code_point(t,ord('d'))
texting.insert_code_point(t,ord('e'))
assert texting.cursor_line_offset(t) == 1
assert texting.cursor_code_point_offset(t) == 2
assert texting.has_been_modified(t) == True
texting.set_unmodified(t)
texting.set_cursor_start(t)
assert texting.current_code_point(t) == ord('a')
texting.advance(t)
assert texting.current_code_point(t) == ord('b')
texting.advance(t)
assert texting.current_code_point(t) == ord('c')
texting.advance(t)
assert texting.current_code_point(t) == ord('\n')
texting.advance(t)
assert texting.current_code_point(t) == ord('d')
texting.advance(t)
assert texting.current_code_point(t) == ord('e')
texting.advance(t)
assert texting.current_code_point(t) == texting.END_OF_TEXT
assert texting.cursor_line_offset(t) == 1
assert texting.cursor_code_point_offset(t) == 2
assert texting.has_been_modified(t) == False
texting.set_cursor(t,0,1)
assert texting.delete_after(t) == True
assert texting.cursor_line_offset(t) == 0
assert texting.cursor_code_point_offset(t) == 1
texting.set_cursor_start(t)
assert texting.current_code_point(t) == ord('a')
texting.advance(t)
assert texting.current_code_point(t) == ord('c')
texting.advance(t)
assert texting.current_code_point(t) == ord('\n')
texting.advance(t)
assert texting.current_code_point(t) == ord('d')
texting.advance(t)
assert texting.current_code_point(t) == ord('e')
texting.advance(t)
assert texting.current_code_point(t) == texting.END_OF_TEXT
texting.advance(t)
assert texting.cursor_line_offset(t) == 1
texting.advance(t)
assert texting.cursor_code_point_offset(t) == 2
texting.advance(t)
assert texting.has_been_modified(t) == True
texting.set_unmodified(t)
texting.set_cursor(t,0,1)
assert texting.delete_after(t) == True
assert texting.delete_after(t) == True
assert texting.cursor_line_offset(t) == 0
assert texting.cursor_code_point_offset(t) == 1
texting.set_cursor_start(t)
assert texting.current_code_point(t) == ord('a')
texting.advance(t)
assert texting.current_code_point(t) == ord('d')
texting.advance(t)
assert texting.current_code_point(t) == ord('e')
texting.advance(t)
assert texting.current_code_point(t) == texting.END_OF_TEXT
assert texting.cursor_line_offset(t) == 0
assert texting.cursor_code_point_offset(t) == 3
assert texting.has_been_modified(t) == True
texting.set_unmodified(t)
texting.set_cursor(t,0,1)
assert texting.delete_after(t) == True
assert texting.delete_after(t) == True
assert texting.delete_after(t) == False
assert texting.delete_after(t) == False
assert texting.cursor_line_offset(t) == 0
assert texting.cursor_code_point_offset(t) == 1
texting.set_cursor_start(t)
assert texting.current_code_point(t) == ord('a')
texting.advance(t)
assert texting.current_code_point(t) == texting.END_OF_TEXT
assert texting.cursor_line_offset(t) == 0
assert texting.cursor_code_point_offset(t) == 1
assert texting.has_been_modified(t) == True
texting.set_unmodified(t)
texting.set_cursor_end(t)
assert texting.delete_after(t) == False
assert texting.delete_after(t) == False
assert texting.delete_after(t) == False
assert texting.delete_after(t) == False
assert texting.cursor_line_offset(t) == 0
assert texting.cursor_code_point_offset(t) == 1
texting.set_cursor_start(t)
assert texting.current_code_point(t) == ord('a')
texting.advance(t)
assert texting.current_code_point(t) == texting.END_OF_TEXT
assert texting.cursor_line_offset(t) == 0
assert texting.cursor_code_point_offset(t) == 1
assert texting.has_been_modified(t) == False
print("OK")

print("")
print("All tests OK")



# Contractor to assist with layout of a page.

# author R.N.Bosworth

# version 13 Mar 23  16:59

from guibits1_0 import font_styling, type_checking2_0, unicoding3_0
from guibits1_0 import windowing, writing
from . import looking_ahead
from . import string_parsing
from . import texting

""" 

Copyright (C) 2019,2021,2022,2023  R.N.Bosworth

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License (gpl.txt) for more details.
"""

# exposed procedures
# ------------------

def width_in_points_of_escaped(win,s,fn,fss,fsize):
  """ 
  pre:
    win = windowing.Window on which s is to be written
    win must be present on the screen
    s = string to be measured, as a unicoding3_0.String
    fn = name of font for s, as a str
    fss = set of font_styling.FontStyles for s (empty set implies the string should be measured plain)
    fsize = point size of font for s, in range 6.0..72.0, as a float 
  post:
    the length of the de-escaped (i.e. rendered) version s in points, for the given window, font name, font styles and font size, has been returned as a float 

  test:
    win = None
    win not showing
    win is showing
      fn = None
        fss= None
          fsize = None
            s = None
      fn = "Courier New"
        fss = {}
          fsize = 5.9
          fsize = 72.1
          fsize = 10.0
            s = ""
            s = "f"
            s = "fred"
            s = "fredfredfredfredfredfredfredfredfredfred"
            s = "1 &lt; 2"
  """
  type_checking2_0.check_derivative(win,windowing.Window)
  type_checking2_0.check_derivative(s,unicoding3_0.String)
  type_checking2_0.check_identical(fn,str)
  type_checking2_0.check_derivative(fss,font_styling.FontStyles)
  type_checking2_0.check_identical(fsize,float)
  if fsize < 6.0 or fsize > 72.0:
    raise Exception("Attempt to lay out font of size <6.0 or >72.0. fsize="+str(fsize))
  ds = unicoding3_0.new_string()
  o = 0
  while o < unicoding3_0.length_of(s):
    (dcp,o) = string_parsing.parse_html_character(s,o)
    unicoding3_0.append(ds,dcp)
  dstr = unicoding3_0.python_string_of(ds)
  return writing.width_in_points_of(win,dstr,fn,fss,fsize)
  

def x_offset_of_line(l,w,a):
  """ 
  pre:
    l = length of the line in points, as float
    w = width of the text on the page which the line is to be displayed, as float
    a = texting.Alignment of the line
    
  post:
    the x-offset from the start of the text area required to display the line 
      with the given alignment has been returned

  test:
    w = None
    w = 5.0
      l = None
      l = 3.0
        a = None
        a = BEGIN
            MIDDLE
            END
            JUSTIFY
  """
  type_checking2_0.check_identical(l,float)
  type_checking2_0.check_identical(w,float)
  type_checking2_0.check_derivative(a,texting.Alignment)
  if a == texting.Alignment.BEGIN:
    return 0.0
  elif a == texting.Alignment.MIDDLE:
    return (w - l)/2.0
  elif a == texting.Alignment.END:
    return w - l
  else:
    pass
  return 0.0



from . import html_emitting, texting
from guibits1_0 import type_checking2_0
from . import unicode_io
from guibits1_0 import unicoding3_0

# author R.N.Bosworth

# version 7 Mar 23  20:14
""" 
Parser which converts an Emily text to an MLE stream.

This parser is written using recursive descent, a la Watt and Brown(2000).

This parser is written as a Contractor, i.e. using pure procedures 
with explicit parameters.

Copyright (C) 2012,2015,2016,2017,2018,2019,2021,2022,2023  R.N.Bosworth

    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License (gpl.txt) for more details.
"""
""" 
  Text Grammar version 4 Feb 2019   09:29
  
  This grammar is in EBNF (Extended Backus-Naur Form) using the following meta-symbols:  ::==  is  (  )  |  *.  The equivalent terminal characters are called LPARENTHESIS, RPARENTHESIS, BAR and STAR.

  Non-terminals are in CamelCase.  Terminal names are in UPPER_CASE.  Other symbols represent themselves.
  
  The rules are case-sensitive.
  
  Text  ::==  OptionalSpaces TextLine TextLine* END_OF_TEXT
  
  TextLine  ::==  Element* NewLine
  
  Element  ::==  Word | Separator

  Word  ::==  NonSpaceChar NonSpaceChar*
  
  NonSpaceChar is any HtmlCharacter except LF, SPACE, TAB, FF or CR
  
  Separator  ::==  SPACE (U+0020)
  
  Newline  ::==  LF (U+000A)
                                                               
  Note 1: Rule for OptionalSpaces is given in HTMLCSSGrammar.  Rule for HtmlCharacter is given in HTMLGrammar.
  
  Note 2: An Internal Paragraph is a series of TextLines terminated by a blank line or END_OF_TEXT.  Every TextLine of an Internal Paragraph has the same alignment.
  ________________________________________________________________
  
  MLE Grammar version 1 Dec 2018   15:14  (reduced version for text translation only)
  
  This grammar is a reduced version of the HTML grammar given in WHATWG's HTML5 website.

  This grammar is in EBNF (Extended Backus-Naur Form) using the following meta-symbols:  ::==  is  (  )  |  *.  The equivalent terminal characters are called LPARENTHESIS, RPARENTHESIS, BAR and STAR.

  Non-terminals are in CamelCase.  Terminal names are in UPPER_CASE.  Other symbols represent themselves.
  
  The rules are case-sensitive, unlike HTML5.
  
  Parameterization is used to distinguish specialized non-terminals such as tags.

  
  BodyElement  ::==  
    StartTag(body) OptionalSpaces (ParagraphOrBreak  OptionalSpaces)* 
	    EndTag(body)
    (note: paragraphs start and end with a Newline in the Emily text)
    
  ParagraphOrBreak  ::==  (Paragraph | BreakTag)
                    
  BreakTag  ::==  StartTag(br)

  Paragraph  ::==  ParagraphStartTag OptionalSpaces ParagraphText EndTag(p)
  
  ParagraphStartTag  ::==  < p ( > | Separator ( > | ClassAttribute OptionalSpaces > ))
  
  ClassAttribute  ::==  c l a s s OptionalSpaces = 
                          OptionalSpaces " ClassName "
  
  ClassName  ::==  ( b e g i n |  m i d d l e | e n d )
                   
  ParagraphText  ::== (HtmlCharacter | BreakTag OptionalSpaces)*
                      (note: in the Emily internal text,
                       code points remain escaped,
                       words are separated by a single space, 
                       and each BreakTag is replaced by a newline.)
                       
  Note: Rules for OptionalSpaces and Separator are given in the HTMLCSSGrammar.  Rules for HtmlCharacter, StartTag and EndTag are given in the HTMLGrammar.
"""

# exposed procedures
# ------------------

def parse_text(t,w):
  """ 
  pre:
    t = texting.Text to be parsed
    w = unicode_io.Writer through which the HTML output is to be written
    
  post:
    text has been emitted as an MLE page (HTML5 standard)

  syntax:
    Text  ::==  OptionalSpaces TextLine TextLine* END_OF_TEXT
    
    TextLine  ::==  Element* NewLine
    
    Element  ::==  Word | Separator

    Word  ::==  NonSpaceChar NonSpaceChar*
    
    NonSpaceChar is any HtmlCharacter except LF, SPACE, TAB, FF or CR
    
    Separator  ::==  SPACE (U+0020)
    
    Newline  ::==  LF (U+000A)

  test:
    w = None
      t = None
    w = valid Writer
      t = ""
      t = "\n\n"
      t = "the"
      t = "\nthe\ncat\n"
          "\nthe\n\ncat\n"
          "\nthe\n\ncat\n\n\nsat\n\n\n\n"
  """
  type_checking2_0.check_derivative(t,texting.Text)
  type_checking2_0.check_derivative(w,unicode_io.Writer)
  loff = texting.cursor_line_offset(t)
  cpoff = texting.cursor_code_point_offset(t)
  texting.set_cursor_start(t)
  in_para = False
  cp = texting.current_code_point(t)
  while cp != texting.END_OF_TEXT:
    if cp == ord('\n'):
      texting.advance(t)
      cp = texting.current_code_point(t)
      if cp == ord('\n') or cp == texting.END_OF_TEXT:
        if not in_para:
          html_emitting.emit_start_tag(unicoding3_0.string_of("br"),w)
          html_emitting.emit_new_line(w)
        else: # in_para
          html_emitting.emit_end_tag(unicoding3_0.string_of("p"),w)
          html_emitting.emit_new_line(w)
          in_para = False
      else:  # NL not followed by NL or EOT
        if in_para:
          html_emitting.emit_start_tag(unicoding3_0.string_of("br"),w)
          html_emitting.emit_new_line(w)
          html_emitting.emit_indent(w)
        else:  # not in_para            
          _emit_paragraph_start_tag(texting.get_alignment(t),w)
          html_emitting.emit_new_line(w)
          html_emitting.emit_indent(w)
          in_para = True
    else:  # not NL
      if not in_para:
        _emit_paragraph_start_tag(texting.get_alignment(t),w)
        html_emitting.emit_new_line(w)
        html_emitting.emit_indent(w)
        in_para = True
      unicode_io.write(cp,w)
      texting.advance(t)
      cp  = texting.current_code_point(t)
  # allow for non-terminated text
  if in_para:
    html_emitting.emit_end_tag(unicoding3_0.string_of("p"),w)
    html_emitting.emit_new_line(w)
    in_para = False
  texting.set_cursor(t,loff,cpoff)
  

# private members
# ---------------

def _emit_paragraph_start_tag(a,w):
  """ 
  pre:
    a = texting.Alignment of paragraph to be emitted
    w = unicode_io.Writer to which paragraph start is to be emitted
    
  post:
    paragraph start tag of Alignment a has been emitted to w

  test:
    a = BEGIN
        MIDDLE
        END
  """
  if a == texting.Alignment.BEGIN:
    html_emitting.emit_start_tag(unicoding3_0.string_of("p class=\"begin\""),w)
  elif a == texting.Alignment.MIDDLE:
    html_emitting.emit_start_tag(unicoding3_0.string_of("p class=\"middle\""),w)
  elif a == texting.Alignment.END:
    html_emitting.emit_start_tag(unicoding3_0.string_of("p class=\"end\""),w)
  else:
    pass



from guibits1_0 import type_checking2_0
from . import unicode_io
from guibits1_0 import unicoding3_0

# author R.N.Bosworth

# version 6 Aug 2022  14:19
""" 
Contractor which allows the emitting of css rules.

Copyright (C) 2016,2017,2021,2022,2023  R.N.Bosworth

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License (gpl.txt) for more details.
"""
""" 
  CSS Grammar version 13.6.17   05:30

  This grammar is a reduced version of the W3C CSS standard, as at April 2016.

  This grammar is in EBNF (Extended Backus-Naur Form) using the following meta-symbols:  ::=  is  (  )  |  *.  The equivalent terminal characters are called LPARENTHESIS, RPARENTHESIS, BAR and STAR.

  Non-terminals are in CamelCase.  Terminal names are in UPPER_CASE.  Other symbols represent themselves.
  
  The rules are case-sensitive, unlike CSS.
    

  StyleSheet  ::=  OptionalSpaces Rule*
  
  Rule  ::=  Selector OptionalSpaces DeclarationBlock OptionalSpaces
  
  Selector  :==  DotName
  
  DotName  ::=  Name ( . Name | )
                   
  DeclarationBlock  ::=  { OptionalSpaces Declaration OptionalSpaces ; OptionalSpaces 
                          (Declaration OptionalSpaces ; OptionalSpaces)* }

  Declaration  ::=  FontDeclaration | TextAlignmentDeclaration
  
  FontDeclaration  ::=  f o n t - (FontNameDeclarationRest | FontSizeDeclarationRest)
    
  FontNameDeclarationRest  ::=  f a m i l y OptionalSpaces : OptionalSpaces FontName
  
  FontName  ::=  " FontNameCharacter FontNameCharacter* "
  
  FontNameCharacter  ::=   A .. Z | a .. z | SPACE
  
  FontSizeDeclarationRest  ::=  s i z e OptionalSpaces : OptionalSpaces Digit ( | Digit) p t
  
  Digit  ::=  0..9
  
  TextAlignmentDeclaration  ::=  t e x t - a l i g n OptionalSpaces : OptionalSpaces TextAlignValue
  
  TextAlignValue  ::=  l e f t | r i g h t | c e n t e r | j u s t i f y

  (Note that "left" and "right" in this context means "the beginning of the line" and "the end of the line", the actual positioning depending on whether the text is read left-to-right or right-to-left.)
  
  Rules for OptionalSpaces and Name are defined in HTMLGrammar.
"""

# exposed procedures
# ------------------

def emit_style_sheet(n,d,w):
  """ 
  pre:
    n = font name for this document's body, as unicoding3_0.String
    d = font size for this document's body in range 0.0 - 99.0, as float
    w = unicode_io.Writer to which this style sheet is to be emitted
    
  post:
    style sheet for this document has been emitted to w,
      including font name and size, 
        and definitions for <p>, <p.begin>, <p.middle>, <p.end> and <p.justify>

  syntax:
    StyleSheet  ::=  OptionalSpaces Rule*
    
    OptionalSpaces  ::=  SpaceChar*

    SpaceChar  ::=  (SPACE | TAB | LF | FF | CR)
                    (U+0020 | U+0009 | U+000A | U+000C | U+000D )
                    
    Rule  ::=  Selector OptionalSpaces DeclarationBlock OptionalSpaces
    
    Selector  :==  DotName
    
    DotName  ::=  Name ( . Name | )

  test:
    once thru
  """
  type_checking2_0.check_derivative(n,unicoding3_0.String)
  type_checking2_0.check_identical(d,float)
  type_checking2_0.check_derivative(w,unicode_io.Writer)
  _emit_font_name_rule(unicoding3_0.string_of("body"),n,w)
  _emit_name_string(_LINE_SEPARATOR,w)
  _emit_font_size_rule(unicoding3_0.string_of("body"),d,w)
  _emit_name_string(_LINE_SEPARATOR,w)
  _emit_text_alignment_rule(unicoding3_0.string_of("p"),unicoding3_0.string_of("left"),w)
  _emit_name_string(_LINE_SEPARATOR,w)
  _emit_text_alignment_rule(unicoding3_0.string_of("p.begin"),unicoding3_0.string_of("left"),w)
  _emit_name_string(_LINE_SEPARATOR,w)
  _emit_text_alignment_rule(unicoding3_0.string_of("p.middle"),unicoding3_0.string_of("center"),w)
  _emit_name_string(_LINE_SEPARATOR,w)
  _emit_text_alignment_rule(unicoding3_0.string_of("p.end"),unicoding3_0.string_of("right"),w)
  _emit_name_string(_LINE_SEPARATOR,w)
  _emit_text_alignment_rule(unicoding3_0.string_of("p.justify"),unicoding3_0.string_of("justify"),w)
  _emit_name_string(_LINE_SEPARATOR,w)
  

# private members
# ---------------
_LINE_SEPARATOR = unicoding3_0.string_of("\n")
# default to Unix

def _emit_font_name_rule(s,n,w):
  """ 
  pre:
    s = selector string for rule, as unicoding3_0.String
    n = font name for rule, as unicoding3_0.String
    w = unicode_io.Writer on which rule is to be emitted
    
  post:
    either:
      appropriate rule has been emitted to w
    or:
      an Exception has been raised

  syntax:
    Rule  ::=  Selector Spaces DeclarationBlock Spaces
    
    Selector  :==  DotName
    
    DotName  ::=  Name ( . Name | )
    
    Name  ::=  NameChar NameChar*

    NameChar  ::=  ( a .. z | 0 .. 9 )
                   ( U+0061 .. U+007A | U+0030 .. U+0039 )
                   
    DeclarationBlock  ::=  { OptionalSpaces Declaration OptionalSpaces ; OptionalSpaces 
                            (Declaration OptionalSpaces ; OptionalSpaces)* }

    Declaration  ::=  FontDeclaration | TextAlignmentDeclaration
    
    FontDeclaration  ::=  f o n t - (FontNameDeclarationRest | FontSizeDeclarationRest)
      
    FontNameDeclarationRest  ::=  f a m i l y OptionalSpaces : OptionalSpaces FontName
    
    FontName  ::=  " FontNameCharacter FontNameCharacter* "
    
    FontNameCharacter  ::=   A .. Z | a .. z | SPACE

  test:
    s = ""
      n = ""
    s = "fred"
      n = "Times New Roman"
  """
  unicode_io.write(ord(' '),w)
  unicode_io.write(ord(' '),w)
  _emit_name_string(s,w)
  unicode_io.write(ord(' '),w)
  unicode_io.write(ord('{'),w)
  unicode_io.write(ord('f'),w)
  unicode_io.write(ord('o'),w)
  unicode_io.write(ord('n'),w)
  unicode_io.write(ord('t'),w)
  unicode_io.write(ord('-'),w)
  unicode_io.write(ord('f'),w)
  unicode_io.write(ord('a'),w)
  unicode_io.write(ord('m'),w)
  unicode_io.write(ord('i'),w)
  unicode_io.write(ord('l'),w)
  unicode_io.write(ord('y'),w)
  unicode_io.write(ord(':'),w)
  unicode_io.write(ord(' '),w)
  unicode_io.write(ord('\"'),w)
  _emit_name_string(n,w)
  unicode_io.write(ord('\"'),w)
  unicode_io.write(ord(';'),w)
  unicode_io.write(ord('}'),w)
  

def _emit_font_size_rule(s,d,w):
  """ 
  pre:
    s = selector string for rule, unicoding3_0.String
    d = font size for rule in range 0.0 - 99.0, as a float
    w = unicode_io.Writer on which rule is to be emitted
    
  post:
    either:
      appropriate rule has been emitted to w
    or:
      IOException has been raised

  syntax:
    Rule  ::=  Selector Spaces DeclarationBlock Spaces
    
    Selector  :==  DotName
    
    DotName  ::=  Name ( . Name | )
    
    Name  ::=  NameChar NameChar*

    NameChar  ::=  ( a .. z | 0 .. 9 )
                   ( U+0061 .. U+007A | U+0030 .. U+0039 )
                   
    FontDeclaration  ::=  f o n t - (FontNameDeclarationRest | FontSizeDeclarationRest)
      
    FontSizeDeclarationRest  ::=  s i z e OptionalSpaces : OptionalSpaces Digit ( | Digit) p t
    
    Digit  ::=  0..9

  test:
    s = ""
      d = 0.0
    s = "jim"
      d = 9.0
    s = "bert"
      d = 10.0
    s = "harold"
      d = 99.0
  """
  unicode_io.write(ord(' '),w)
  unicode_io.write(ord(' '),w)
  _emit_name_string(s,w)
  unicode_io.write(ord(' '),w)
  unicode_io.write(ord('{'),w)
  unicode_io.write(ord('f'),w)
  unicode_io.write(ord('o'),w)
  unicode_io.write(ord('n'),w)
  unicode_io.write(ord('t'),w)
  unicode_io.write(ord('-'),w)
  unicode_io.write(ord('s'),w)
  unicode_io.write(ord('i'),w)
  unicode_io.write(ord('z'),w)
  unicode_io.write(ord('e'),w)
  unicode_io.write(ord(':'),w)
  unicode_io.write(ord(' '),w)
  i = int(d)
  dig0 = i%10
  dig1 = int(i/10)
  if dig1 != 0:
    unicode_io.write(ord('0') + dig1,w)
  unicode_io.write(ord('0') + dig0,w)
  unicode_io.write(ord('p'),w)
  unicode_io.write(ord('t'),w)
  unicode_io.write(ord(';'),w)
  unicode_io.write(ord('}'),w)
  

def _emit_name_string(s,w):
  """ 
  pre:
    s = name string which is to be emitted, as unicoding3_0.String
          (string must be lower case and must not contain '<' or '&')
    w = unicode_io.Writer to which s is to be emitted
    
  post:
    s has been emitted on w

  syntax:
    DotName  ::=  Name ( . Name | )
    
    Name  ::=  NameChar NameChar*

    NameChar  ::=  ( a .. z | 0 .. 9 )
                   ( U+0061 .. U+007A | U+0030 .. U+0039 )

  test:
    empty string
    "cat"
  """
  i = 0
  while i < unicoding3_0.length_of(s):
    unicode_io.write(unicoding3_0.code_point_at(s,i),w)    
    i += 1
  

def _emit_text_alignment_rule(s,a,w):
  """ 
  pre:
    s = selector string for rule, as unicoding3_0.String
    a = alignment value for this selector, unicoding3_0.String
    w = unicode_io.Writer on which rule is to be emitted
    
  post:
    either:
      appropriate rule has been emitted to w
    or:
      an Exception has been raised

  syntax:
    Rule  ::=  Selector OptionalSpaces DeclarationBlock OptionalSpaces
    
    Selector  :==  DotName
    
    DotName  ::=  Name ( . Name | )
                     
    Name  ::=  NameChar NameChar*

    NameChar  ::=  ( a .. z | 0 .. 9 )
                   ( U+0061 .. U+007A | U+0030 .. U+0039 )

    DeclarationBlock  ::=  { OptionalSpaces Declaration OptionalSpaces ; OptionalSpaces 
                            (Declaration OptionalSpaces ; OptionalSpaces)* }

    Declaration  ::=  FontDeclaration | TextAlignmentDeclaration
    
    TextAlignmentDeclaration  ::=  t e x t - a l i g n OptionalSpaces : OptionalSpaces TextAlignValue
    
    TextAlignValue  ::=  l e f t | r i g h t | c e n t e r | j u s t i f y

    (Note that "left" and "right" in this context means "the beginning of the line" and "the end of the line", the actual positioning depending on whether the text is read left-to-right or right-to-left.)
    
  test:
    s = ""
      a = ""
    s = "fred"
      a = "left"
  """
  unicode_io.write(ord(' '),w)
  unicode_io.write(ord(' '),w)
  _emit_name_string(s,w)
  unicode_io.write(ord(' '),w)
  unicode_io.write(ord('{'),w)
  unicode_io.write(ord('t'),w)
  unicode_io.write(ord('e'),w)
  unicode_io.write(ord('x'),w)
  unicode_io.write(ord('t'),w)
  unicode_io.write(ord('-'),w)
  unicode_io.write(ord('a'),w)
  unicode_io.write(ord('l'),w)
  unicode_io.write(ord('i'),w)
  unicode_io.write(ord('g'),w)
  unicode_io.write(ord('n'),w)
  unicode_io.write(ord(':'),w)
  unicode_io.write(ord(' '),w)
  _emit_name_string(a,w)
  unicode_io.write(ord(';'),w)
  unicode_io.write(ord('}'),w)
  

# test program
# ------------

def _test():
  print("Tests of _emit_name_string", end=' ')
  w = unicode_io.new_string_writer()
  s = unicoding3_0.new_string()
  _emit_name_string(s,w)
  assert unicoding3_0.length_of(unicode_io.get_string(w)) == 0
  w = unicode_io.new_string_writer()
  s = unicoding3_0.string_of("cat")
  _emit_name_string(s,w)
  assert unicoding3_0.equals(unicode_io.get_string(w),unicoding3_0.string_of("cat"))
  print("OK")
  
  print("Tests of _emit_font_name_rule", end=' ')
  w = unicode_io.new_string_writer()
  s = unicoding3_0.new_string()
  n = unicoding3_0.new_string()
  _emit_font_name_rule(s,n,w)
  assert unicoding3_0.equals(unicode_io.get_string(w),unicoding3_0.string_of("   {font-family: \"\";}"))
  s = unicoding3_0.string_of("fred")
  n = unicoding3_0.string_of("Times New Roman")
  w = unicode_io.new_string_writer()
  _emit_font_name_rule(s,n,w)
  assert unicoding3_0.equals(unicode_io.get_string(w),unicoding3_0.string_of("  fred {font-family: \"Times New Roman\";}"))
  print("OK")
  
  print("Tests of _emit_font_size_rule", end=' ')
  s = unicoding3_0.new_string()
  w = unicode_io.new_string_writer()
  _emit_font_size_rule(s,0.0,w)
  assert unicoding3_0.equals(unicode_io.get_string(w),unicoding3_0.string_of("   {font-size: 0pt;}"))
  s = unicoding3_0.string_of("jim")
  w = unicode_io.new_string_writer()
  _emit_font_size_rule(s,9.0,w)
  assert unicoding3_0.equals(unicode_io.get_string(w),unicoding3_0.string_of("  jim {font-size: 9pt;}"))
  s = unicoding3_0.string_of("bert")
  w = unicode_io.new_string_writer()
  _emit_font_size_rule(s,10.0,w)
  assert unicoding3_0.equals(unicode_io.get_string(w),unicoding3_0.string_of("  bert {font-size: 10pt;}"))
  s = unicoding3_0.string_of("harold")
  w = unicode_io.new_string_writer()
  _emit_font_size_rule(s,99.0,w)
  assert unicoding3_0.equals(unicode_io.get_string(w),unicoding3_0.string_of("  harold {font-size: 99pt;}"))
  print("OK")
  
  print("Tests of _emit_text_alignment_rule", end=' ')
  s = unicoding3_0.new_string()
  a = unicoding3_0.new_string()
  w = unicode_io.new_string_writer()
  _emit_text_alignment_rule(s,a,w)
  assert unicoding3_0.equals(unicode_io.get_string(w),unicoding3_0.string_of("   {text-align: ;}"))
  s = unicoding3_0.string_of("fred")
  a = unicoding3_0.string_of("left")
  w = unicode_io.new_string_writer()
  _emit_text_alignment_rule(s,a,w)
  assert unicoding3_0.equals(unicode_io.get_string(w),unicoding3_0.string_of("  fred {text-align: left;}"))
  print("OK")
  
  print("Tests of emit_style_sheet", end=' ')
  w = unicode_io.new_output_writer("css.test")
  emit_style_sheet(unicoding3_0.string_of("Palatino"),18.0,w)
  unicode_io.write(unicode_io.END_OF_STREAM,w)
  print("  check file css.test")
  print("OK")
  
  print("")


if __name__ == "__main__":
  import sys
  _test()
  print("All tests OK")



""" 
Contractor for Unicode input and output.
Input and output is performed on a codepoint-by-codepoint basis,
with unicode_io.END_OF_STREAM as stream terminator.
Externally, the Unicode stream is stored in UTF-8 format.
It is also possible to input and output to a unicoding3_0.String,
for testing purposes.
"""

# author R.N.Bosworth

# version 8 Mar 23  11:11

from guibits1_0 import type_checking2_0, unicoding3_0

"""
Copyright (C) 2017,2021,2022  R.N.Bosworth

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License (gpl.txt) for more details.
"""

# exposed constants
# -----------------

END_OF_STREAM =  - 1


# exposed types
# -------------

class Reader:
  pass
  

class Writer:
  pass
  

# exposed procedures
# ------------------

def get_string(w):
  """ 
  pre:
    w = _StringWriter from which string is to be extracted
    
  post:
    w's string has been returned

  test:
    w is not a _StringWriter
    w is a _StringWriter with a non-empty string
  """
  type_checking2_0.check_derivative(w,_StringWriter)
  return w.get_string()
  

def new_input_reader(fn):
  """ 
  pre:
    fn = filename from which the Unicode stream is to be read, as a str
    
  post:
    either:
      an input Reader object has been returned for the specified filename
    or:
      an Exception has been raised

  test:
    invalid file name (invalid characters)
    valid file name but file not there
    valid existent file
  """
  type_checking2_0.check_derivative(fn,str)
  r = _InputReader()
  try:
    r._my_file = open(fn)
  except:
    raise Exception("Unable to open file:" + fn)
  return r
  

def new_output_writer(fn):
  """ 
  pre:
    fn = filename to which the Unicode stream is to be written, as str
    
  post:
    either:
      an output Writer object has been returned for the specified filename
    or:
      an Exception has been raised

  test:
    file name not str
    invalid file name (invalid characters)
    valid file name but file not there (remove hardware)
    valid existent file
  """
  type_checking2_0.check_identical(fn,str)
  try:
    w = _OutputWriter()
    w._my_file = open(fn,'w')
  except:
    raise Exception("Output file could not be opened")
  return w
  
  
def new_string_reader(s):
  """ 
  pre:
    s = unicoding3_0.String from which the the reader is to be created
    
  post:
    a string Reader object has been returned for the specified string

  test:
    s = None
    s = invalid string
    s = valid unicoding3_0.String
  """
  type_checking2_0.check_derivative(s,unicoding3_0.String)
  sr = _StringReader()
  sr._my_string = s
  sr._pos = 0
  return sr
  

def new_string_writer():
  """ 
  post:
    a string Writer object has been returned

  test:
    once thru
  """
  sw = _StringWriter()
  sw._my_string = unicoding3_0.new_string()
  return sw
    

def read(r):
  """ 
  pre:
    r = Reader from which the next code point is to be read
    
  post:
    either:
      the next code point, or END_OF_STREAM, has been returned from r
    or:
      an IOException has been raised

  test:
    _Reader:
      'a'
      _READER_EOS
    _StringReader:
      U+0000
      "abc"
      U+FFFF
      U+10000
      U+10FFFF
  """
  type_checking2_0.check_derivative(r,Reader)
  return r.read()
  

def write(cp,w):
  """ 
  pre:
    cp = next Unicode code point to be written, or END_OF_STREAM
    w = Writer to which cp is to be written
    
  post:
    either:
      the next code point, or END_OF_STREAM, has been written to w
    or:
      an IOException has been raised

  test:
    cp = -2
    cp = -1
    cp = 0x10ffff
    cp = 0x110000
  """
  type_checking2_0.check_identical(cp,int)
  if cp < END_OF_STREAM or cp > 0x10ffff:
    raise Exception("Attempt to write invalid code point: "+ hex(cp))
  type_checking2_0.check_derivative(w,Writer)
  w.write(cp)
  

# private classes
# ---------------

class _InputReader(Reader):

  def read(self):
    """ 
    pre:
      self = _InputReader from which the next code point is to be read
      
    post:
      either:
        the next code point, or END_OF_STREAM, has been returned from r
      or:
        an IOException has been raised

    test:
      U+0000
      'a'
      U+FFFF
      U+10000
      U+10FFFF
      _READER_EOS
    """
    c = self._my_file.read(1)
    if c == '':
      return END_OF_STREAM
    else:
      return ord(c)
      

class _OutputWriter(Writer):

  def write(self,cp):      
    """ 
    pre:
      self = _OutputWriter on which cp is to be written
      cp = next code point to be written, or END_OF_STREAM
      
    post:
      either:
        the next code point, or END_OF_STREAM, has been written to the underlying stream
      or:
        an IOException has been raised

    test:
      0x0000
      'a'
      0xffff
      0x10000
      0x10ffff
      END_OF_STREAM
    """
    if cp == END_OF_STREAM:
      self._my_file.close()
    else:
      self._my_file.write(chr(cp))    
  
  def get_string(self):
    """ 
    pre:
      self = Writer from which string is required
    post:
      a RuntimeException has been raised
      
    test:
      once thru
    """
    raise Exception("Attempt to get string from a non-string Writer")


class _StringReader(Reader):

  def read(self):
    """
    pre:
      self = Reader from which the next code point is to be read
      
    post:
      the next code point, or END_OF_STREAM, has been returned
      
    test:
      not END_OF_STREAM
      END_OF_STREAM
    """
    if self._pos >= unicoding3_0.length_of(self._my_string):
      return END_OF_STREAM
      
    else:
      cp = unicoding3_0.code_point_at(self._my_string,self._pos)
      self._pos += 1
      return cp


class _StringWriter(Writer):

  def write(self,cp):
    """ 
    pre:
      self = Writer to which cp is to be written
      cp = next code point to be written, or END_OF_STREAM
      
    post:
      the next code point has been written to the underlying string
        (END_OF_STREAM is ignored)

    test:
      'x'
        'y'
          'z'
            END_OF_STREAM
    """
    if cp != END_OF_STREAM:
      unicoding3_0.append(self._my_string,cp)
      

  def get_string(self):
    """ 
    pre:
      self = _StringWriter from which string is to be extracted
      
    post:
      this writer's string has been returned

    test:
      empty string
      non-empty string
    """
    return self._my_string



# Test program for 0ne symbol lookahead stream.

# author R.N.Bosworth

# version 4 Mar 23  16:05

from emily0_9 import looking_ahead
from guibits1_0 import unicoding3_0

"""
Copyright (C) 2012,2015,2016,2017,2018,2021,2022,2023  R.N.Bosworth

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License (gpl.txt) for more details.
"""

# test program
# ------------

print("Tests of file instantiation and reading", end=' ')
#l = looking_ahead.lookahead_of("motorbike?$&*")
#l = looking_ahead.lookahead_of("motorbike")
l = looking_ahead.lookahead_of("motorbike.test")
assert looking_ahead.current_symbol_of(l) == ord('x')
assert looking_ahead.current_symbol_of(l) == ord('x')
looking_ahead.advance(l)
assert looking_ahead.current_symbol_of(l) == ord('y')
looking_ahead.advance(l)
assert looking_ahead.current_symbol_of(l) == ord('z')
looking_ahead.advance(l)
assert looking_ahead.current_symbol_of(l) == looking_ahead.END_OF_STREAM
print("OK")

print("Tests of string instantiation and reading", end=' ')
l = looking_ahead.lookahead_of_string(unicoding3_0.string_of("abc"))
assert looking_ahead.current_symbol_of(l) == ord('a')
assert looking_ahead.current_symbol_of(l) == ord('a')
looking_ahead.advance(l)
assert looking_ahead.current_symbol_of(l) == ord('b')
looking_ahead.advance(l)
assert looking_ahead.current_symbol_of(l) == ord('c')
looking_ahead.advance(l)
assert looking_ahead.current_symbol_of(l) == looking_ahead.END_OF_STREAM
print("OK")

print("")
print("All tests OK")



# Test program for Contractor to assist with layout of a page.

# author R.N.Bosworth

# version 13 Mar 23  16:54

from emily0_9 import page_laying_out, texting
from guibits1_0 import font_styling, windowing, unicoding3_0

""" 
Copyright (C) 2019,2021,2022,2023  R.N.Bosworth

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License (gpl.txt) for more details.
"""

# test program
# ------------

def _equals_float(d1,d2):
  """ 
  pre:
    d1 = first floating-point number
    d2 = second floating-point number
    
  post:
    returns true iff the two floating-point numbers are within 0.01 of each other

  test:
    d1 = 123.4
      d2 = 123.3    
      d2 = 123.401    
      d2 = 123.5    
  """
  return abs(d1 - d2) < 0.01
  

def window_closing(win):
  #page_laying_out.width_in_points_of_escaped(win,None,None,None,None)
  #page_laying_out.width_in_points_of_escaped(win,unicoding3_0.string_of(""),None,None,None)
  #page_laying_out.width_in_points_of_escaped(win,unicoding3_0.string_of(""),"Courier New",None,None)
  fss = font_styling.new_font_styles()
  #page_laying_out.width_in_points_of_escaped(win,unicoding3_0.string_of(""),"Courier New",fss,None)
  #page_laying_out.width_in_points_of_escaped(win,unicoding3_0.string_of(""),"Courier New",fss,5.9)
  #page_laying_out.width_in_points_of_escaped(win,unicoding3_0.string_of(""),"Courier New",fss,72.1)
  assert page_laying_out.width_in_points_of_escaped(win,unicoding3_0.string_of(""),"Courier New",fss,10.0) == 0.0
  wipoe = page_laying_out.width_in_points_of_escaped(win,unicoding3_0.string_of("f"),"Courier New",fss,10.0)
  assert _equals_float(wipoe,6.0)
  wipoe = page_laying_out.width_in_points_of_escaped(win,unicoding3_0.string_of("fred"),"Courier New",fss,10.0) 
  assert _equals_float(wipoe,24.0)
  wipoe = page_laying_out.width_in_points_of_escaped(win,unicoding3_0.string_of("fredfredfredfredfredfredfredfredfredfred"),"Courier New",fss,10.0) 
  assert _equals_float(wipoe,240.04)
  wipoe = page_laying_out.width_in_points_of_escaped(win,unicoding3_0.string_of("1 &lt; 2"),"Courier New",fss,10.0) 
  assert _equals_float(wipoe,30.0)
  print("OK")
  print("")
  print("All tests OK")
  return True

print("Tests of page_laying_out.x_offset_of_line", end=' ')
#page_laying_out.x_offset_of_line(None,None,None)
#page_laying_out.x_offset_of_line(3.0,None,None)
#page_laying_out.x_offset_of_line(3.0,5.0,None)
assert page_laying_out.x_offset_of_line(3.0,5.0,texting.Alignment.BEGIN) == 0.0
assert page_laying_out.x_offset_of_line(3.0,5.0,texting.Alignment.MIDDLE) == 1.0
assert page_laying_out.x_offset_of_line(3.0,5.0,texting.Alignment.END) == 2.0
print("OK")


print("Tests of page_laying_out.width_in_points_of_escaped", end=' ')
#page_laying_out.width_in_points_of_escaped(None,None,None,None,None)
win = windowing.new_window(12.0,"Window for test of page_laying_out.width_in_points_of_escaped",800.0,500.0,1.0)
fss = font_styling.new_font_styles()
#assert page_laying_out.width_in_points_of_escaped(win,unicoding3_0.string_of(""),"Courier New",fss,12.0) == 0.0
windowing.show(win,None,window_closing)
print("OK")

print("")
print("All tests OK")



# test of main class subprocedures for Emily

# author R.N.Bosworth

# version 16 Mar 23  22:05
"""
Copyright (C) 2015,2016,2017,2018,2019,2021,2022,2023  R.N.Bosworth

  This program is free software: you can redistribute it and/or modify
  it under the terms of the GNU General Public License as published by
  the Free Software Foundation, either version 3 of the License, or
  (at your option) any later version.

  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License (gpl.txt) for more details.
"""

from emily0_9 import emily, persisting

#------------------------------------------------------------------------------
# test program
# -----------------------------------------------------------------------------

def _print_text(t):
  cpo = texting.cursor_code_point_offset(t)
  lo = texting.cursor_line_offset(t)
  print("-----------------------------------")
  texting.set_cursor_start(t)
  cp = texting.current_code_point(t)
  texting.advance(t)
  while cp != texting.END_OF_TEXT:
    print("" + (chr(cp)))
    cp = texting.current_code_point(t)
    texting.advance(t)
    
  print("-----------------------------------")
  texting.set_cursor(t,lo,cpo)
  
  
print("Tests of emily.build_tagged_label", end=' ')
r = emily.build_tagged_label("fred",False)
assert r == "   fred"
r = emily.build_tagged_label("fred",True)
assert r == "• fred"
print("OK")

print ("Tests of build_tagged_font_size_label", end=' ')
l = emily.build_tagged_font_size_label(10,11)
assert l == "   10 point"  
l = emily.build_tagged_font_size_label(11,11)
assert l == "• 11 point"  
print("OK")

print("Tests of check_file_type")
emily.my_persistents = persisting.get_persistents()
fn = emily.check_file_type("fred.mlf")
print("  fn="+str(fn))
fn = emily.check_file_type("fred.mle")
print("  fn="+str(fn))
print("OK")

print("Tests of int_of",end=' ')
#i = emily.int_of("")
#i = emily.int_of("fred")
i = emily.int_of("  90")
assert i == 90
i = emily.int_of("  909 points")
assert i == 909
print("OK")

print("Tests of name_of",end=' ')
#s = emily.name_of("")
#s = emily.name_of("£££££")
assert emily.name_of("fredfont") == "fredfont"
assert emily.name_of("& Times New Roman") == "Times New Roman"
print("OK")

print("")
print("All tests OK")



#Test program for Contractor for emitting an MLE document.

# author R.N.Bosworth

# version 9 Mar 23  15:24

from emily0_9 import mle_emitting, texting, unicode_io
from guibits1_0 import unicoding3_0

"""
Copyright (C) 2017,2018,2021,2022,2023  R.N.Bosworth

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License (gpl.txt) for more details.
"""

# test program
# ------------

print("Tests of emit_emily_document", end=' ')
#mle_emitting.emit_emily_document(None,None,None,None,None)
n = unicoding3_0.string_of("Liberty")
#mle_emitting.emit_emily_document(n,None,None,None,None)
#mle_emitting.emit_emily_document(n,-0.1,None,None,None)
#mle_emitting.emit_emily_document(n,99.1,None,None,None)
d = 10.0
#mle_emitting.emit_emily_document(n,d,None,None,None)
v = unicoding3_0.string_of("99.9")
#mle_emitting.emit_emily_document(n,d,v,None,None)
t = texting.new_text()
#mle_emitting.emit_emily_document(n,d,v,t,None)
w = unicode_io.new_output_writer("test.mle")
mle_emitting.emit_emily_document(n,d,v,t,w)
texting.insert_code_point(t,ord('<'))
texting.insert_code_point(t,ord('a'))
texting.insert_code_point(t,ord('l'))
texting.insert_code_point(t,ord('i'))
texting.insert_code_point(t,ord('g'))
texting.insert_code_point(t,ord('n'))
texting.insert_code_point(t,ord('b'))
texting.insert_code_point(t,ord('e'))
texting.insert_code_point(t,ord('g'))
texting.insert_code_point(t,ord('i'))
texting.insert_code_point(t,ord('n'))
texting.insert_code_point(t,ord('>'))
texting.insert_code_point(t,ord('h'))
texting.insert_code_point(t,ord('e'))
texting.insert_code_point(t,ord('l'))
texting.insert_code_point(t,ord('l'))
texting.insert_code_point(t,ord('o'))
texting.insert_code_point(t,ord('\n'))
texting.insert_code_point(t,ord('\n'))
w = unicode_io.new_output_writer("test2.mle")
mle_emitting.emit_emily_document(n,d,v,t,w)
print("OK")

print("")
print("All tests OK")



from . import code_point_testing
from . import css_parsing
from guibits1_0 import font_styling
from . import html_css_parsing
from . import html_parsing
from . import looking_ahead
from . import texting
from guibits1_0 import type_checking2_0
from . import unicode_io
from guibits1_0 import unicoding3_0

# author R.N.Bosworth

# version 13 Mar 23  12:54
""" 
Contractor which parses an mle file via a code point lookahead to produce an Emily text.

This parser is written using recursive descent, a la Watt and Brown(2000).

Copyright (C) 2012,2015,2016,2017,2018,2019,2021,2023  R.N.Bosworth

    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License (gpl.txt) for more details.
"""
""" 
  MLE Grammar version 1 Dec 2018   15:14
  
  This grammar is a reduced version of the HTML grammar given in WHATWG's HTML5 website.

  This grammar is in EBNF (Extended Backus-Naur Form) using the following meta-symbols:  ::==  is  (  )  |  *.  The equivalent terminal characters are called LPARENTHESIS, RPARENTHESIS, BAR and STAR.

  Non-terminals are in CamelCase.  Terminal names are in UPPER_CASE.  Other symbols represent themselves.
  
  The rules are case-sensitive, unlike HTML5.
  
  Parameterization is used to distinguish specialized non-terminals such as tags.

  
  MleDocument  ::==  OptionalSpaces Doctype OptionalSpaces HtmlElement OptionalSpaces

  Doctype  ::==  < ! d o c t y p e Separator h t m l OptionalSpaces >
  
  HtmlElement  ::==  StartTag(html) OptionalSpaces HeadElement OptionalSpaces BodyElement 
                      OptionalSpaces EndTag(html)

  HeadElement  ::==  StartTag(head) OptionalSpaces MetaTag OptionalSpaces TitleElement 
                      OptionalSpaces StyleElement OptionalSpaces EndTag(head)
                     
  MetaTag  ::==  < m e t a Separator c h a r s e t OptionalSpaces = OptionalSpaces " U T F - 8 " 
                  OptionalSpaces >

  TitleElement  ::==  StartTag(title) Text EndTag(title)

  Text  ::==  HtmlCharacter*

  StyleElement  ::==  StartTag(style) OptionalSpaces StyleSheet OptionalSpaces EndTag(style)
                      (for StyleSheet syntax see CSSGrammar.txt)

  BodyElement  ::==  
    StartTag(body) OptionalSpaces (ParagraphOrBreak  OptionalSpaces)* 
	    EndTag(body)
    (note: paragraphs are separated by a blank line in the Emily text)
    
  ParagraphOrBreak  ::==  (Paragraph | BreakTag)
                    
  BreakTag  ::==  StartTag(br)

  Paragraph  ::==  ParagraphStartTag OptionalSpaces ParagraphText EndTag(p)
  
  ParagraphStartTag  ::==  < p ( > | Separator ( > | ClassAttribute OptionalSpaces > )) (ci)
  
  ClassAttribute  ::==  c l a s s OptionalSpaces = 
                          OptionalSpaces " ClassName "
  
  ClassName  ::==  ( b e g i n |  m i d d l e | e n d )
                   
  ParagraphText  ::== (HtmlCharacter | BreakTag OptionalSpaces)*
                      (note: in the Emily internal text,
                       code points remain escaped,
                       words are separated by a single space, 
                       and each BreakTag is replaced by a newline.)
                       
  Note: Rules for OptionalSpaces and Separator are given in the HTMLCSSGrammar.  Rules for HtmlCharacter, StartTag and EndTag are given in the HTMLGrammar.
   
  Simplifications from HTML5:
  
   1. No legacy string in Doctype.
   
   2. Html start and end tags mandatory.
   
   3. No attributes in start tag (for the moment).
   
   4. No solidus in start tag.
   
   5. Head contains only title.
   
   6. Body contains only paragraphs and break tags.
   
   7. Paragraph has mandatory end tag.
   
   8. Comments not allowed (for security).
   
   9. Tags and attribute names are case-sensitive.
"""
""" 
  Text Grammar version 4 Feb 2019   09:29
  
  This grammar is in EBNF (Extended Backus-Naur Form) using the following meta-symbols:  ::==  is  (  )  |  *.  The equivalent terminal characters are called LPARENTHESIS, RPARENTHESIS, BAR and STAR.

  Non-terminals are in CamelCase.  Terminal names are in UPPER_CASE.  Other symbols represent themselves.
  
  The rules are case-sensitive.
  
  Text  ::==  OptionalSpaces TextLine TextLine* END_OF_TEXT
  
  TextLine  ::==  Element* NewLine
  
  Element  ::==  Word | Separator

  Word  ::==  NonSpaceChar NonSpaceChar*
  
  NonSpaceChar is any HtmlCharacter except LF, SPACE, TAB, FF or CR
  
  Separator  ::==  SPACE (U+0020)
  
  Newline  ::==  LF (U+000A)
                                                               
  Note 1: Rule for OptionalSpaces is given in HTMLCSSGrammar.  Rule for HtmlCharacter is given in HTMLGrammar.
  
  Note 2: An Internal Paragraph is a series of TextLines terminated by a blank line or END_OF_TEXT.  Every TextLine of an Internal Paragraph has the same alignment.
"""

# exposed types
# -------------

class MLEResult:
  pass
  

# exposed procedures
# ------------------

def alignment_of(r,s):
  """ 
  pre:
    r = MLEResult of an mle parse
    s = string whose alignment is to be found
  
  post:
    alignment corresponding to s from r has been returned

  test:
    once thru
  """
  type_checking2_0.check_derivative(r,MLEResult)
  type_checking2_0.check_derivative(s,unicoding3_0.String)
  return css_parsing.alignment_of(r._my_style_properties,s)
  

def font_name_of(r):
  """ 
  pre:
    r = MLEResult of an mle parse
  
  post:
    font name discovered in parse has been returned

  test:
    once thru
  """
  type_checking2_0.check_derivative(r,MLEResult)
  return css_parsing.font_name_of(r._my_style_properties)
  

def font_size_of(r):
  """ 
  pre:
    r = MLEResult of an mle parse
  
  post:
    font size discovered in parse has been returned

  test:
    once thru
  """
  type_checking2_0.check_derivative(r,MLEResult)
  return css_parsing.font_size_of(r._my_style_properties)
  

def parse_mle_document(la):
  """ 
  pre:
    la = lookahead to be parsed
  
  post:
    either:
      Emily document has been parsed from la and the MLEResult returned
    or:
      Exception has been thrown
  
  syntax:
    MLEDocument  ::=  OptionalSpaces Doctype OptionalSpaces HtmlElement OptionalSpaces

  test:
    "<!doctype word><html><head><title>Your Grace</title><style></style></head><body><p>A word.</p></body></html>"
    "<!doctype html><itml><head><title>Your Grace</title><style></style></head><body><p>A word.</p></body></html>"
    "<!doctype html><html><head><title>Your Grace</title><style></style></head><body><p>A word.</p></body></html>"
    "  <!doctype html>  <html>  <head>  <title>Your Grace</title>  <style>  p  { font-family  :  \"Helvetica\"  ;  }  p  {  font-size  :  2pt  ;  }  </style>  </head>  <body>  <p>A word.</p>  </body>  </html>  "
  """
  type_checking2_0.check_derivative(la,looking_ahead.Lookahead)
  html_css_parsing.accept_optional_spaces(la)
  _parse_doctype(la)
  html_css_parsing.accept_optional_spaces(la)
  r = _parse_html_element(la)
  html_css_parsing.accept_optional_spaces(la)
  html_css_parsing.accept(looking_ahead.END_OF_STREAM,la)
  return r
  

def text_of(r):
  """ 
  pre:
    r = MLEResult of an mle parse
  
  post:
    text discovered in parse has been returned

  test:
    once thru
  """
  type_checking2_0.check_derivative(r,MLEResult)
  return r._my_text
  

# private procedures
# ------------------

def _emit_hard_line_break(t):
  """ 
  pre:
    t = text whose content is to be extended, with cursor at end
  
  post:
    a hard line break has been emitted to t

  test:
    once thru
  """
  texting.insert_code_point(t,ord('\n'))


def _new_mle_result():
  """
  pre:
    True
    
  post:
    new MLEResult variable has been returned, with default values
    
  test:
    once thru (check values)
  """
  mler = MLEResult()
  mler._my_style_properties = None
  mler._my_text = None
  return mler
  

def _parse_body_element(la,t):
  """ 
  pre:
    la = lookahead to be parsed
    t = text whose content is to be extended, with cursor at end
  
  post:
    either:
      Emily body element has been parsed from la and appended as text to t
    or:
      RuntimeException has been thrown

  syntax:
    BodyElement  ::=  
      StartTag(body) OptionalSpaces (ParagraphOrBreak  OptionalSpaces)* 
        EndTag(body)
      (note: paragraphs are separated by a blank line in the Emily text)

  test:
    "<cody>  </body>"
    "<body>  </aody>"
    "<body>  </body>"
    "<body>  <br>  <p class = "middle">   first  para  </p>  <br>  <p>  1 &lt; 2 &amp; 2 &lt; 3  </p>  <br>  </body>"
  """
  html_parsing.accept_start_tag(unicoding3_0.string_of("body"),la)
  html_css_parsing.accept_optional_spaces(la)
  end_of_body = _parse_paragraph_or_break(la,t)
  while  not end_of_body:
    html_css_parsing.accept_optional_spaces(la)
    end_of_body = _parse_paragraph_or_break(la,t)
    
  
def _parse_class_attribute(la):
  """ 
  pre:
    la = lookahead to be parsed
  
  post:
    either:
      ClassAttribute has been parsed
      name of class element has been returned
    or:
      RuntimeException has been thrown

  syntax:
    ClassAttribute  ::==  c l a s s OptionalSpaces = 
                            OptionalSpaces " ClassName "

  tests:
    form="middle"
    class   -   "middle"
    class   =   'middle'
    class   =   "moddle"
    class   =   "middle'
    class   =   "middle"
    class="begin"
  """
  html_css_parsing.accept_name(unicoding3_0.string_of("class"),la)
  html_css_parsing. accept_optional_spaces(la)
  html_css_parsing.accept(ord('='),la)
  html_css_parsing.accept_optional_spaces(la)
  html_css_parsing.accept(ord('\"'),la)
  s = _parse_class_name(la)
  html_css_parsing.accept(ord('\"'),la)
  return s
  

def _parse_class_name(la):
  """ 
  pre:
    la = lookahead to be parsed
  
  post:
    either:
      ClassName has been parsed
      class name has been returned
    or:
      RuntimeException has been thrown

  syntax:
    ClassName  ::==  ( b e g i n |  m i d d l e | e n d ) 

  tests:
    "start"
    "begin"
    "middle"
    "end"
  """
  s = html_css_parsing.parse_name(la)
  if unicoding3_0.equals(s,unicoding3_0.string_of("begin")) or unicoding3_0.equals(s,unicoding3_0.string_of("middle")) or unicoding3_0.equals(s,unicoding3_0.string_of("end")) or unicoding3_0.equals(s,unicoding3_0.string_of("justify")):
    return s
  else:
    raise Exception("Unexpected class name: \"" + unicoding3_0.python_string_of(s) + "\"")
    
  
def _parse_doctype(la):
  """ 
  pre:
    la = lookahead to be parsed
  
  post:
    either:
      Doctype has been parsed
    or:
      RuntimeException has been thrown

  syntax:
    Doctype  ::=  < ! d o c t y p e Separator h t m l OptionalSpaces >  !!

  test:
    ">!doctype html>"
    "<£doctype html>"
    "<!doctypf html>"
    "<!doctypehtml>"
    "<!doctype   htmm>"
    "<!doctype   html   <"
    "<!doctype   html   >"
    "<!doctype html>"
  """
  html_css_parsing.accept(ord('<'),la)
  html_css_parsing.accept(ord('!'),la)
  html_css_parsing.accept_name(unicoding3_0.string_of("doctype"),la)
  html_css_parsing.accept_separator(la)
  html_css_parsing.accept_name(unicoding3_0.string_of("html"),la)
  html_css_parsing.accept_optional_spaces(la)
  html_css_parsing.accept(ord('>'),la)
  

def _parse_head_element(la):
  """ 
  pre:
    la = token lookahead to be parsed
  
  post:
    either:
      Emily HeadElement has been parsed from la
      StyleProperties object for this HeadElement has been returned
    or:
      a RuntimeException has been thrown

  syntax:
    HeadElement  ::=  StartTag(head) OptionalSpaces MetaTag OptionalSpaces TitleElement 
                        OptionalSpaces StyleElement OptionalSpaces EndTag(head)

    TitleElement  ::=  StartTag(title) Text EndTag(title)

    Text  ::=  EscapedUnicodeCodePoint*

    StyleElement  ::=  StartTag(style) OptionalSpaces StyleSheet OptionalSpaces EndTag(style)
                        (for StyleSheet syntax see CSSGrammar.txt)

  test:
    "<title></title>
    "<head></head>"
    "<head>  </head>"
    "<head>  <meta charset=\"UTF-7\">"
    "<head>  <meta charset=\"UTF-8\">"
    "<head>  <meta charset=\"UTF-8\">  <title>"
    "<head>  <meta charset=\"UTF-8\">  <title>  </head>"
    "<head>  <meta charset=\"UTF-8\">  <title></title>  </title>"
    "<head>  <meta charset=\"UTF-8\">  <title></title>  </head>"
    "<head>  <meta charset=\"UTF-8\">  <title></title>  <stylf>  </style>  </head>"
    "<head>  <meta charset=\"UTF-8\">  <title></title>  <style>  </styld>  </head>"
    "<head>  <meta charset=\"UTF-8\">  <title>alchemist of the year</title>  <style>  </style>  </head>"
    "<head>  <meta charset=\"UTF-8\">  <title>fred</title>  <style>  p  {font-size:66pt;}  </style>  </head>"
  """
  html_parsing.accept_start_tag(unicoding3_0.string_of("head"),la)
  html_css_parsing.accept_optional_spaces(la)
  _parse_meta_tag(la)
  html_css_parsing.accept_optional_spaces(la)
  html_parsing.accept_start_tag(unicoding3_0.string_of("title"),la)
  while looking_ahead.current_symbol_of(la) != ord('<') and looking_ahead.current_symbol_of(la) != looking_ahead.END_OF_STREAM:
    looking_ahead.advance(la)
  html_parsing.accept_end_tag(unicoding3_0.string_of("title"),la)
  html_css_parsing.accept_optional_spaces(la)
  html_parsing.accept_start_tag(unicoding3_0.string_of("style"),la)
  sp = css_parsing.parse_style_sheet(la)
  html_parsing.accept_end_tag(unicoding3_0.string_of("style"),la)
  html_css_parsing.accept_optional_spaces(la)
  html_parsing.accept_end_tag(unicoding3_0.string_of("head"),la)
  return sp
  

def _parse_html_element(la):
  """
  pre:
    la = token lookahead to be parsed
    w = window on which Emily text is to be displayed
    mlw = maximum line width in points
    mph = maximum page height in points
  
  post:
    either:
      Emily HTML element has been parsed from la and an MLEResult has been returned
    or:
      a RuntimeException has been thrown

  syntax:
    HtmlElement  ::=  StartTag(html) OptionalSpaces HeadElement OptionalSpaces
                        BodyElement OptionalSpaces EndTag(html)

  test:
    "<htmm><head><title>Your Grace</title><style></style></head><body><p>A word.</p></body></html>"
    "<html><head><title>Your Grace</title><style></style></head><body><p>A word.</p></body></htmk>"
    "<html><head><title>Your Grace</title><style></style></head><body><p>A word.</p></body></html>"
    "<html><head><title>Your Grace</title><style>p{font-size:42pt;}</style></head><body><p>A word.</p></body></html>"
    "  <html>  <head>  <title>Your Grace</title>  <style>  p  {font-size:  53pt  ;  }  </style>  </head>  <body>  <p>  A phrase.  </p>  </body>  </html>  "
  """
  html_parsing.accept_start_tag(unicoding3_0.string_of("html"),la)
  html_css_parsing.accept_optional_spaces(la)
  sp = _parse_head_element(la)
  html_css_parsing.accept_optional_spaces(la)
  fss = font_styling.new_font_styles()
  t = texting.new_text()
  _parse_body_element(la,t)
  html_css_parsing.accept_optional_spaces(la)
  html_parsing.accept_end_tag(unicoding3_0.string_of("html"),la)
  mler = _new_mle_result()
  mler._my_style_properties = sp
  mler._my_text = t
  return mler
  

def _parse_meta_tag(la):
  """ 
  pre:
    la = lookahead to be parsed
  
  post:
    either:
      MetaTag has been parsed
    or:
      RuntimeException has been thrown
  
  syntax:

    MetaTag  ::=  < m e t a Separator c h a r s e t OptionalSpaces = OptionalSpaces 
                    " U T F - 8 " OptionalSpaces >
  test:
    ">meta charset=\"UTF-8\">"
    "<beta charset=\"UTF-8\">"
    "<metacharset=\"UTF-8\">"
    "<meta   charabanc=\"UTF-8\">"
    "<meta charset   -    \"UTF-8\">"
    "<meta charset   =    \'UTF-8\">"
    "<meta charset   =    \"UTE-8\">"
    "<meta charset   =    \"UTF-8\'>"
    "<meta charset   =    \"UTF-8\"<"
    "<meta charset   =    \"utf-8\">"
    "<meta charset=\"UTF-8\">"
  """
  html_css_parsing.accept(ord('<'),la)
  html_css_parsing.accept_name(unicoding3_0.string_of("meta"),la)
  html_css_parsing.accept_separator(la)
  html_css_parsing.accept_name(unicoding3_0.string_of("charset"),la)
  html_css_parsing.accept_optional_spaces(la)
  html_css_parsing.accept(ord('='),la)
  html_css_parsing.accept_optional_spaces(la)
  html_css_parsing.accept(ord('\"'),la)
  html_css_parsing.accept(ord('U'),la)
  html_css_parsing.accept(ord('T'),la)
  html_css_parsing.accept(ord('F'),la)
  html_css_parsing.accept(ord('-'),la)
  html_css_parsing.accept(ord('8'),la)
  html_css_parsing.accept(ord('\"'),la)
  html_css_parsing.accept(ord('>'),la)
  

def _parse_paragraph_or_break(la,t):
  """ 
  pre:
    la = token lookahead to be parsed
    t = text whose content is to be extended, with cursor at end
  
  post:
    either:
      a paragraph or a break tag has been parsed from la into t
      "false" has been returned
    or:
      the end of the body has been reached, 
        the </body> tag has been accepted, 
      "true" has been returned
    or:
      a RuntimeException has been thrown
  
  syntax:
    ParagraphOrBreak  ::=  (Paragraph | BreakTag)
                      
    BreakTag  ::=  StartTag(br)

    Paragraph  ::=  ParagraphStartTag OptionalSpaces ParagraphText EndTag(p)

    ParagraphStartTag  ::=  < p ( > | Separator ( > | ClassAttribute OptionalSpaces > ))

  tests:
    >p>
    <q>
    <bt>
    <br   <
    <br   >
    <p></p>
    <p>hello</q>
    <p>hello</p>
    <p   >hello</p>
    <p   clunk>hello</p>
    <p   class   =   "middle"   <
    <p   class   =   "justify"   >hello</p>
    <pclass="middle">hello</p>
    <p class="end">hello<br>world</p>
    <p class="end">hello</p><br>world
    <p class="middle">hello</p>
    <p class="muddle">hello</p>
    <p   class   =   "begin"   >   hello   </p   >
    </cody>
    </body>
  """
  paragraph_alignment = texting.Alignment.BEGIN
  html_css_parsing.accept(ord('<'),la)
  if looking_ahead.current_symbol_of(la) == ord('b'):
    looking_ahead.advance(la)
    html_css_parsing.accept(ord('r'),la)
    html_css_parsing.accept_optional_spaces(la)
    html_css_parsing.accept(ord('>'),la)
    _emit_hard_line_break(t)
    texting.set_alignment(t,texting.Alignment.BEGIN)
    return False
  elif looking_ahead.current_symbol_of(la) == ord('p'):
    looking_ahead.advance(la)
    if looking_ahead.current_symbol_of(la) == ord('>'):
      looking_ahead.advance(la)
      paragraph_alignment = texting.Alignment.BEGIN
      texting.set_alignment(t,paragraph_alignment)
    else:
      html_css_parsing.accept_separator(la)
      if looking_ahead.current_symbol_of(la) == ord('>'):
        looking_ahead.advance(la)
        paragraph_alignment = texting.Alignment.BEGIN
        texting.set_alignment(t,paragraph_alignment)
      else:
        class_attribute = _parse_class_attribute(la)
        html_css_parsing.accept_optional_spaces(la)
        html_css_parsing.accept(ord('>'),la)
        if unicoding3_0.equals(class_attribute,unicoding3_0.string_of("begin")):
          paragraph_alignment = texting.Alignment.BEGIN
        elif unicoding3_0.equals(class_attribute,unicoding3_0.string_of("middle")):
          paragraph_alignment = texting.Alignment.MIDDLE
        elif unicoding3_0.equals(class_attribute,unicoding3_0.string_of("end")):
          paragraph_alignment = texting.Alignment.END
        else:
          raise Exception("Unexpected paragraph class attribute: \"" + unicoding3_0.python_string_of(class_attribute))
        texting.set_alignment(t,paragraph_alignment)
  elif looking_ahead.current_symbol_of(la) == ord('/'):
    looking_ahead.advance(la)
    html_css_parsing.accept_name(unicoding3_0.string_of("body"),la)
    html_css_parsing.accept(ord('>'),la)
    return True
  else:
    raise Exception("Unexpected tag label start in MLE body: expected \'p\' or \'b\', found \'" + chr(looking_ahead.current_symbol_of(la)) + "\'")
  _emit_hard_line_break(t)
  texting.set_alignment(t,paragraph_alignment)
  html_css_parsing.accept_optional_spaces(la)
  tagname = _parse_paragraph_text(la,paragraph_alignment,t)
  if not unicoding3_0.equals(tagname,unicoding3_0.string_of("/p")):
    raise Exception("Unexpected tag at end of paragraph: expected </p>, found <" + unicoding3_0.python_string_of(tagname) + ">")
  else:
    _emit_hard_line_break(t)
    texting.set_alignment(t,paragraph_alignment)
  return False
  

def _parse_paragraph_text(la,pa,t):
  """ 
  pre:
    la = token lookahead to be parsed
    pa = alignment of this paragraph: BEGIN, MIDDLE or END
    t = text whose content is to be extended, with cursor at end
  
  post:
    either:
      paragraph text been parsed from la and appended as text to t
      the name of the current tag (following the text) has been returned
    or:
      a RuntimeException has been thrown

  syntax:
    ParagraphText  ::= (EscapedUnicodeCodePoint | BreakTag OptionalSpaces)*
                        (note: in the Emily internal text,
                         code points remain escaped,
                         words are separated by a single space, 
                         and each BreakTag is translated to a newline.)
                         
      BreakTag  ::=  StartTag(br)

  test:
    pa = "middle"
      </p>
      <br><bq>
      <br>a<bs>
      a<br>&lt;b</p>
      ab</p>
      a b</p>
      a  b</p>
      a \t b   cd   </p>
  """
  current_tag = unicoding3_0.new_string()
  is_start_of_line = True
  is_start_of_word = True
  end_of_paragraph_text = False

  while  not end_of_paragraph_text:
    cp = looking_ahead.current_symbol_of(la)
    if cp == ord('<'):
      current_tag = html_parsing.parse_tag(la)
      if unicoding3_0.equals(current_tag,unicoding3_0.string_of("br")):
        texting.set_alignment(t,pa)
        _emit_hard_line_break(t)
        texting.set_alignment(t,pa)
        html_css_parsing.accept_optional_spaces(la)
        is_start_of_line = True
        is_start_of_word = True

      # not a BreakTag, must be end of paragraph text
      else:
        end_of_paragraph_text = True

    # not a tag
    else:
      if code_point_testing.is_space_char(cp):
        html_css_parsing.accept_optional_spaces(la)
        is_start_of_word = True
        
      # not a space char
      else:
        if is_start_of_word:
          if  not is_start_of_line:
            texting.insert_code_point(t,ord(' '))
          is_start_of_line = False
          is_start_of_word = False
        texting.insert_code_point(t,cp)
        looking_ahead.advance(la)
    
  return current_tag



# Test of Parser which converts an Emily text to an MLE stream.

# author R.N.Bosworth

# version 7 Mar 23  20:10

from emily0_9 import texting, text_emitting, unicode_io
from guibits1_0 import unicoding3_0

""" 
Copyright (C) 2012,2015,2016,2017,2018,2019,2021,2022,2023  R.N.Bosworth

    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License (gpl.txt) for more details.
"""

# test program
# ------------

print("Tests of _emit_paragraph_start_tag", end=' ')
w = unicode_io.new_string_writer()
text_emitting._emit_paragraph_start_tag(texting.Alignment.BEGIN,w)
assert unicoding3_0.equals(unicode_io.get_string(w),unicoding3_0.string_of("<p class=\"begin\">"))
w = unicode_io.new_string_writer()
text_emitting._emit_paragraph_start_tag(texting.Alignment.MIDDLE,w)
assert unicoding3_0.equals(unicode_io.get_string(w),unicoding3_0.string_of("<p class=\"middle\">"))
w = unicode_io.new_string_writer()
text_emitting._emit_paragraph_start_tag(texting.Alignment.END,w)
assert unicoding3_0.equals(unicode_io.get_string(w),unicoding3_0.string_of("<p class=\"end\">"))
print("OK")

print("Tests of parse_text", end=' ')
#text_emitting.parse_text(None,None)
t = texting.new_text()
#text_emitting.parse_text(t,None)
w = unicode_io.new_string_writer()
text_emitting.parse_text(t,w)
assert unicoding3_0.equals(unicode_io.get_string(w),unicoding3_0.string_of(""))
texting.insert_code_point(t,ord('\n'))
texting.insert_code_point(t,ord('\n'))
text_emitting.parse_text(t,w)
assert unicoding3_0.equals(unicode_io.get_string(w),unicoding3_0.string_of("<br>\n<br>\n"))
t = texting.new_text()
texting.insert_code_point(t,ord('t'))
texting.insert_code_point(t,ord('h'))
texting.insert_code_point(t,ord('e'))
texting.set_cursor_start(t)
w = unicode_io.new_string_writer()
text_emitting.parse_text(t,w)
assert unicoding3_0.equals(unicode_io.get_string(w),unicoding3_0.string_of("<p class=\"begin\">\n  the</p>\n"))
t = texting.new_text()
texting.insert_code_point(t,ord('\n'))
texting.insert_code_point(t,ord('t'))
texting.insert_code_point(t,ord('h'))
texting.insert_code_point(t,ord('e'))
texting.insert_code_point(t,ord('\n'))
texting.insert_code_point(t,ord('c'))
texting.insert_code_point(t,ord('a'))
texting.insert_code_point(t,ord('t'))
texting.insert_code_point(t,ord('\n'))
w = unicode_io.new_string_writer()
text_emitting.parse_text(t,w)
assert unicoding3_0.equals(unicode_io.get_string(w),unicoding3_0.string_of("<p class=\"begin\">\n  the<br>\n  cat</p>\n"))
t = texting.new_text()
texting.insert_code_point(t,ord('\n'))
texting.insert_code_point(t,ord('t'))
texting.insert_code_point(t,ord('h'))
texting.insert_code_point(t,ord('e'))
texting.insert_code_point(t,ord('\n'))
texting.insert_code_point(t,ord('\n'))
texting.insert_code_point(t,ord('c'))
texting.insert_code_point(t,ord('a'))
texting.insert_code_point(t,ord('t'))
texting.insert_code_point(t,ord('\n'))
w = unicode_io.new_string_writer()
text_emitting.parse_text(t,w)
assert unicoding3_0.equals(unicode_io.get_string(w),unicoding3_0.string_of("<p class=\"begin\">\n  the</p>\n<p class=\"begin\">\n  cat</p>\n"))
t = texting.new_text()
texting.insert_code_point(t,ord('\n'))
texting.insert_code_point(t,ord('t'))
texting.insert_code_point(t,ord('h'))
texting.insert_code_point(t,ord('e'))
texting.insert_code_point(t,ord('\n'))
texting.insert_code_point(t,ord('\n'))
texting.insert_code_point(t,ord('c'))
texting.insert_code_point(t,ord('a'))
texting.insert_code_point(t,ord('t'))
texting.insert_code_point(t,ord('\n'))
texting.insert_code_point(t,ord('\n'))
texting.insert_code_point(t,ord('\n'))
texting.insert_code_point(t,ord('s'))
texting.insert_code_point(t,ord('a'))
texting.insert_code_point(t,ord('t'))
texting.insert_code_point(t,ord('\n'))
texting.insert_code_point(t,ord('\n'))
texting.insert_code_point(t,ord('\n'))
texting.insert_code_point(t,ord('\n'))
w = unicode_io.new_string_writer()
text_emitting.parse_text(t,w)
assert unicoding3_0.equals(unicode_io.get_string(w),unicoding3_0.string_of("<p class=\"begin\">\n  the</p>\n<p class=\"begin\">\n  cat</p>\n<br>\n<p class=\"begin\">\n  sat</p>\n<br>\n<br>\n<br>\n"))
print("OK")

print("")
print("All tests OK")



# Tests of contractor for testing Unicode code points.

# author R.N.Bosworth

# version 2 Mar 2023  14:50

"""
Copyright (C) 2018,2021,2022,2023  R.N.Bosworth

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License (gpl.txt) for more details.
"""
# test program
# ------------

from emily0_9 import code_point_testing

print("Tests of code_point_testing.is_name_char", end=' ')
#code_point_testing.is_name_char('@')
assert  not code_point_testing.is_name_char(ord('@'))
assert  not code_point_testing.is_name_char(ord('A'))
assert  not code_point_testing.is_name_char(ord('Z'))
assert  not code_point_testing.is_name_char(ord('['))
assert  not code_point_testing.is_name_char(ord('`'))
assert code_point_testing.is_name_char(ord('a'))
assert code_point_testing.is_name_char(ord('z'))
assert  not code_point_testing.is_name_char(ord('{'))
assert  not code_point_testing.is_name_char(ord('/'))
assert code_point_testing.is_name_char(ord('0'))
assert code_point_testing.is_name_char(ord('9'))
assert  not code_point_testing.is_name_char(ord(':'))
print("OK")

print("Tests of code_point_testing.is_space_char", end=' ')
#code_point_testing.is_space_char(' ')
assert code_point_testing.is_space_char(ord(' '))
assert code_point_testing.is_space_char(ord('\t'))
assert code_point_testing.is_space_char(ord('\n'))
assert code_point_testing.is_space_char(ord('\f'))
assert code_point_testing.is_space_char(ord('\r'))
assert  not code_point_testing.is_space_char(ord('!'))
print("OK")

print("")
print("All tests OK")



# Test program for Contractor to parse an HTML string.

# author R.N.Bosworth

# version 4 Mar 23  16:27

from emily0_9 import string_parsing
from guibits1_0 import unicoding3_0

""" 
Copyright (C) 2019,2021,2022,2023  R.N.Bosworth

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License (gpl.txt) for more details.
"""

# test program
# ------------

print("Tests of string_parsing._accept", end=' ')
cp = ord('A')
s = unicoding3_0.string_of("B")
#o = string_parsing._accept(cp,s,0);
s = unicoding3_0.string_of("A")
o = string_parsing._accept(cp,s,0)
assert o == 1
print("OK")

print("Tests of string_parsing.parse_html_character", end=' ')
#(cp,o) = string_parsing.parse_html_character(None,None)
s = unicoding3_0.string_of("&bmp;")
#(cp,o) = string_parsing.parse_html_character(s,None)
#(cp,o) =string_parsing.parse_html_character(s,0)
s = unicoding3_0.string_of("&anp;")
#(cp,o) = string_parsing.parse_html_character(s,0)
s = unicoding3_0.string_of("&amq;")
#(cp,o) = string_parsing.parse_html_character(s,0)
s = unicoding3_0.string_of("&amp;")
(cp,o) = string_parsing.parse_html_character(s,0)
assert cp == ord('&')
assert o == unicoding3_0.length_of(s)
s = unicoding3_0.string_of("&ft;")
#(cp,o) = string_parsing.parse_html_character(s,0)
s = unicoding3_0.string_of("&gs;")
#(cp,o) = string_parsing.parse_html_character(s,0)
s = unicoding3_0.string_of("&gt;")
(cp,o) = string_parsing.parse_html_character(s,0)
assert cp == ord('>')
assert o == unicoding3_0.length_of(s)
s = unicoding3_0.string_of("&mt;")
#(cp,o) = string_parsing.parse_html_character(s,0)
s = unicoding3_0.string_of("&lu;")
#(cp,o) = string_parsing.parse_html_character(s,0)
s = unicoding3_0.string_of("&lt;")
(cp,o) = string_parsing.parse_html_character(s,0)
assert cp == ord('<')
assert o == unicoding3_0.length_of(s)
s = unicoding3_0.string_of(">")
#(cp,o) = string_parsing.parse_html_character(s,0)
s = unicoding3_0.string_of("<")
#(cp,o) = string_parsing.parse_html_character(s,0)
s = unicoding3_0.string_of("a")
(cp,o) = string_parsing.parse_html_character(s,0)
assert cp == ord('a')
assert o == unicoding3_0.length_of(s)
print("OK")

print("")
print("All tests OK")



# Contractor for testing Unicode code points.

# author R.N.Bosworth

# version 25 Feb 2023  19:20

from guibits1_0 import type_checking2_0

"""
Contractor for testing Unicode code points.

Copyright (C) 2018,2021,2022,2023  R.N.Bosworth

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License (gpl.txt) for more details.
"""
"""
  HTML CSS Grammar version 17.3.18   15:39

  This grammar defines the basic HTML constructs common to HTML5 and CSS.

  This grammar is in EBNF (Extended Backus-Naur Form) using the following meta-symbols:  ::=  is  (  )  |  *.  The equivalent terminal characters are called LPARENTHESIS, RPARENTHESIS, BAR and STAR.

  Non-terminals are in CamelCase.  Terminal names are in UPPER_CASE.  Other symbols represent themselves.
  
  The rules are case-sensitive, unlike HTML5.
  
  
  OptionalSpaces  ::=  SpaceChar*

  SpaceChar  ::=  (SPACE | TAB | LF | FF | CR)
                  (U+0020 | U+0009 | U+000A | U+000C | U+000D )
                  
  Separator  ::==  SpaceChar OptionalSpaces
  
  Name  ::=  NameChar NameChar*

  NameChar  ::=  ( a .. z | 0 .. 9 )
                 ( U+0061 .. U+007A | U+0030 .. U+0039 )
                 
"""

# exposed procedures
# ------------------


def is_name_char(cp):
  """
  pre:
    cp = code point to be tested
  
  post:
    returns true iff cp is an HTML NameChar
  """
  """
  NameChar  ::=  ( a .. z | 0 .. 9 )
                 ( U+0061 .. U+007A | U+0030 .. U+0039 )  !!
  """
  """
  test:
    cp is not an int
    cp = ord('@')
              A
              Z
              [
              U+0060
              a
              z
              {
              /
              0
              9
              :         
  """
  type_checking2_0.check_identical(cp,int)
  return (ord('a') <= cp and cp <= ord('z') or ord('0') <= cp and cp <= ord('9'))
  

def is_space_char(cp):
  """
  pre:
    cp = code point to be tested

  post:
    returns true iff cp is an HTML SpaceChar
  """
  """
  SpaceChar  <-  ( SPACE | TAB | LF | FF | CR )
               ( U+0020 | U+0009 | U+000A | U+000C | U+000D )
  """
  """
  test:
    cp is not an int
    cp = SPACE
         TAB
         LF
         FF
         CR
         ord('!')
  """
  type_checking2_0.check_identical(cp,int)
  return (cp == ord(' ') or cp == ord('\t') or cp == ord('\n') or cp == ord('\f') or cp == ord('\r'))
  


# Test of Contractor which allows the emitting of css rules.

from emily0_9 import css_emitting, unicode_io
from guibits1_0 import unicoding3_0

# author R.N.Bosworth

# version 1 Mar 23  16:04 

# test program
# ------------

print("Tests of css_emitting._emit_name_string", end=' ')
w = unicode_io.new_string_writer()
s = unicoding3_0.new_string()
css_emitting._emit_name_string(s,w)
assert unicoding3_0.length_of(unicode_io.get_string(w)) == 0
w = unicode_io.new_string_writer()
s = unicoding3_0.string_of("cat")
css_emitting._emit_name_string(s,w)
assert unicoding3_0.equals(unicode_io.get_string(w),unicoding3_0.string_of("cat"))
print("OK")

print("Tests of css_emitting._emit_font_name_rule", end=' ')
w = unicode_io.new_string_writer()
s = unicoding3_0.new_string()
n = unicoding3_0.new_string()
css_emitting._emit_font_name_rule(s,n,w)
assert unicoding3_0.equals(unicode_io.get_string(w),unicoding3_0.string_of("   {font-family: \"\";}"))
s = unicoding3_0.string_of("fred")
n = unicoding3_0.string_of("Times New Roman")
w = unicode_io.new_string_writer()
css_emitting._emit_font_name_rule(s,n,w)
assert unicoding3_0.equals(unicode_io.get_string(w),unicoding3_0.string_of("  fred {font-family: \"Times New Roman\";}"))
print("OK")

print("Tests of css_emitting._emit_font_size_rule", end=' ')
s = unicoding3_0.new_string()
w = unicode_io.new_string_writer()
css_emitting._emit_font_size_rule(s,0.0,w)
assert unicoding3_0.equals(unicode_io.get_string(w),unicoding3_0.string_of("   {font-size: 0pt;}"))
s = unicoding3_0.string_of("jim")
w = unicode_io.new_string_writer()
css_emitting._emit_font_size_rule(s,9.0,w)
assert unicoding3_0.equals(unicode_io.get_string(w),unicoding3_0.string_of("  jim {font-size: 9pt;}"))
s = unicoding3_0.string_of("bert")
w = unicode_io.new_string_writer()
css_emitting._emit_font_size_rule(s,10.0,w)
assert unicoding3_0.equals(unicode_io.get_string(w),unicoding3_0.string_of("  bert {font-size: 10pt;}"))
s = unicoding3_0.string_of("harold")
w = unicode_io.new_string_writer()
css_emitting._emit_font_size_rule(s,99.0,w)
assert unicoding3_0.equals(unicode_io.get_string(w),unicoding3_0.string_of("  harold {font-size: 99pt;}"))
print("OK")

print("Tests of css_emitting._emit_text_alignment_rule", end=' ')
s = unicoding3_0.new_string()
a = unicoding3_0.new_string()
w = unicode_io.new_string_writer()
css_emitting._emit_text_alignment_rule(s,a,w)
assert unicoding3_0.equals(unicode_io.get_string(w),unicoding3_0.string_of("   {text-align: ;}"))
s = unicoding3_0.string_of("fred")
a = unicoding3_0.string_of("left")
w = unicode_io.new_string_writer()
css_emitting._emit_text_alignment_rule(s,a,w)
assert unicoding3_0.equals(unicode_io.get_string(w),unicoding3_0.string_of("  fred {text-align: left;}"))
print("OK")

print("Tests of css_emitting.emit_style_sheet")
w = unicode_io.new_output_writer("css.test")
css_emitting.emit_style_sheet(unicoding3_0.string_of("Palatino"),18.0,w)
unicode_io.write(unicode_io.END_OF_STREAM,w)
print("  check file css.test")
print("OK")

print("")
print("All tests OK")
